"""
    Define various processes for the Portable Bridge Notation (PBN) format.

    This handles BfG specific items, e.g. the "Bids" token.
    (Standard PBN has the "Auction" Token.)
"""

import os
from termcolor import cprint

from datetime import datetime
from bridgeobjects import Hand, Call, Auction, SEATS
from bfgsupport import Board
from ..source.player import Player

PBN_VERSION = '2.1'
CREATOR = 'Bid for Game'

MODULE_COLOUR = 'cyan'

class PBN(object):
    """Create a PBN object."""
    SUITS = {'NT': -1, 'S': 0, 'H': 1, 'D': 2, 'C': 3, -1: 'NT', 0: 'S', 1: 'H', 2: 'D', 3: 'C'}

    def __init__(self, event='', site='', date=''):
        self.event = event
        self.site = site
        if not date:
            date = datetime.now()
        self.date = date
        self.board_count = 0

    def boards_from_path(self, path):
        """Return a list of boards given the path to the .pbn file."""
        boards = []
        if path:
            if os.path.isfile(path):
                with open(path, 'r') as f_event:
                    event_lines = f_event.readlines()
                    boards = self.deal_list(event_lines)
        return boards

    def deal_list(self, file_lines, north_first_hand=True):
        """Return a list of BfG Deals generated from the file_lines list."""
        test_boards = []
        board = None
        board_description = ''
        for row in file_lines:
            text_in_row = self._text_in_row(row)
            if not text_in_row:
                if board:
                    board.description = board_description
                    test_boards.append(board)
            elif row.startswith('[Board'):
                description = row[len('Board')+3:-3]
                board_description = description
            elif row.startswith('[Dealer'):
                dealer = row[len('Dealer')+3:-3]
            elif row.startswith('[Deal'):
                board = self._get_board_from_row(row, north_first_hand)
            if board:
                if row.startswith('[Bids'):
                    board.auction = Auction()
                    bids = row[len('Bids')+3:-3]
                    bid_list = bids.split(' ')
                    board.auction.calls = [Call(call) for call in bid_list]
        if board:
            board.description = board_description
            test_boards.append(board)
        return test_boards

    def _get_board_from_row(self, row, north_first_hand):
        """Return a Board object from a string."""
        deal_string = row[7:-3]
        dealer = deal_string[0][0]
        #######
        if north_first_hand:
            norths_hand = 0
        else:
            norths_hand = SEATS.index(dealer)
        #######
        raw_hands = deal_string[2:].split(" ")
        board = Board()
        board.dealer = dealer
        for hand_string in raw_hands:
            hand_list = self._pbn_hand_list(hand_string)
            hand = Hand(hand_list)

            board.hands[norths_hand] = hand
            board.players[norths_hand] = Player(board, hand)

            norths_hand += 1
            norths_hand %= 4
        return board

    @staticmethod
    def _text_in_row(row):
        """
        Return the text in a string after the removal of 'semi-space' characters.
        """
        delete_chars = [9, 32, 13, 10]
        text_in_row = row
        for delete_char in delete_chars:
            text_in_row = text_in_row.replace(chr(delete_char), "")
        return text_in_row

    def _pbn_hand_list(self, hand_string):
        """Return a list for each hand in bfg format."""
        hand_list = []
        hand_cards = hand_string.split(".")
        for index, suit_cards in enumerate(hand_cards):
            suit = self.SUITS[index]
            for card in suit_cards:
                card_name = card+suit
                hand_list.append(card_name)
        return hand_list

    def save_boards(self, boards, path):
        """Save a list of boards to the path."""
        file_rows = self.get_file_rows(boards)
        rows = "\n".join(file_rows)
        with open(path, 'w') as f_pbn_event:
            f_pbn_event.writelines(rows)

    def get_file_rows(self, boards):
        """Return a list of strings in PBN format generated by a list of boards."""
        file_rows = []
        first_row = True
        for board in boards:
            file_rows.extend(self._get_board_rows(board, first_row))
            first_row = False
        return file_rows

    @staticmethod
    def _get_header_rows():
        """Return a list of strings in PBN format header."""
        header_rows = [f'% PBN {PBN_VERSION}',
                       '% EXPORT',
                       '% Content-type: text/pbn; charset=ISO-8859-1',
                       f'% Creator: {CREATOR}',
                       '%']
        return header_rows

    def _get_board_rows(self, board, first_row=False):
        """Return a list of strings in PBN format generated by a board."""
        self.board_count += 1
        board_rows = []
        if first_row:
            board_rows.append(f'[Event "{self.event}"]')
            board_rows.append(f'[Site "{self.site}"]')
            board_rows.append(f'[Date "{self.date:%Y.%m.%d}"]')
        board_rows.append(f'[Board "{self.board_count} {board.identifier}"]')
        board_rows.append(f'[Dealer "{board.dealer}"]')
        board_rows.append(f'[Deal "{board.dealer}:{self._get_pbn_hands(board)}"]')
        board_rows.append('')
        return board_rows

    def _get_pbn_hands(self, board):
        """Return a board as a string in pbn format."""
        hand_list = []
        offset = SEATS.index(board.dealer)

        for _ in range(4):
            hand = board.hands[offset]
            cards = self._card_values_from_hand(hand)
            suit_list = self._suit_list_from_card_list(cards)
            hand_list.append('.'.join(suit_list))
            offset += 1
            offset %= 4
        board_string = ' '.join(hand_list)
        return board_string

    def _card_values_from_hand(self, hand):
        """
        Return a list of card values by suit for a hand.
        e.g. 8S, 7S, 5S, 9H, 6H ... converts to
        [['8', '7', '5'], ['9', '6'] ...
        """
        cards = [[], [], [], []]
        for card in hand.cards:
            card_value = card.name[0]
            suit = card.name[1]
            cards[self.SUITS[suit]].append(card_value)
        return cards

    @staticmethod
    def _suit_list_from_card_list(card_list):
        """
        Return a list of cards in suits from the card_list.
        e.g. [['8', '7', '5'], ['9', '6'] ... ... converts to
        ['875', '96', ...]
        """
        suit_list = []
        for index in range(4):
            suit_list.append(''.join(card_list[index]))
        return suit_list



""" Bid for Game
    Acol RespondersLaterBid module
"""
import inspect
from .bidding import Bid, Pass
from .blackwood import Blackwood
from .hand import Hand


class RespondersLaterBid(Hand):
    """BfG RespondersLaterBid class."""
    def __init__(self, hand_cards, board):
        super(RespondersLaterBid, self).__init__(hand_cards, board)
        self.openers_last_bid = Bid(self.bid_history[-2])
        self.openers_penultimate_bid = Bid(self.bid_history[-6])
        self.trace = 0

    def suggested_bid(self):
        """Direct control to relevant method and return a Bid object."""
        suit = self.openers_last_bid.denomination

        if self.openers_penultimate_bid.name == '5NT':
            bid = Pass('0522')
        elif self.opener_bid_one.name == '2C':
            bid = self._two_club_opening()
        elif self.openers_last_bid.is_nt and self.openers_last_bid.level >= 4:
            bid = self._opener_bid_blackwood()
        elif self.previous_bid.is_nt and self.previous_bid.level >= 4:
            bid = self._responder_bid_nt()
        elif self._opener_bids_three_nt_after_change_of_major():
            bid = self.next_level_bid(self.opener_bid_one.denomination, '0529')
        elif self._opener_bids_three_nt_and_hand_has_six_card_major():
            bid = self.suit_bid(4, self.longest_suit, '0205')
        elif self._opener_rebids_second_suit_at_three_level():
            bid = self.suit_bid(4, self.bid_two.denomination, '0524')
        elif self._can_support_openers_second_suit_and_twelve_points():
            bid = self.next_level_bid(suit, '0525')
        elif self._opener_jumped_and_can_support():
            bid = self.nt_bid(4, '0526')
        elif self._opener_bid_minor_at_level_three_twelve_points():
            bid = self.nt_bid(3, '0527')
        elif self._opener_rebids_first_suit_two_card_support():
            bid = self.next_level_bid(self.opener_bid_one.denomination, '0528')
        elif self._has_support_for_openers_new_suit_at_level_four():
            bid = self.next_level_bid(self.opener_bid_three.denomination, '0465')
        elif self.hcp >= 16 and self.nt_level <= 4:
            bid = self.nt_bid(4, '0079')
        elif self._opener_strong_and_balanced_and_fifteen_points():
            bid = self.nt_bid(6, '0202')
        elif self._two_card_support_for_openers_six_card_suit_twelve_points():
            bid = self.next_level_bid(self.opener_bid_one.denomination, '0251')
        elif self._opener_support_six_card_suit() and not Bid(self.bid_history[-2]).is_game:
            bid = self.next_level_bid(self.longest_suit, '0343')
        elif self._has_support_for_openers_six_card_suit():
            bid = self.next_level_bid(self.opener_bid_one.denomination, '0541')
        elif self._opener_at_level_three_nine_points_and_stoppers():
            bid = self.nt_bid(3, '0286')
        elif self._opener_repeats_second_suit():
            bid = self._opener_has_rebid_second_suit()
        elif self.openers_last_bid.denomination == self.longest_suit and self.openers_last_bid.is_game:
            bid = Pass('0935')
        elif self._support_after_opener_shows_six_four():
            bid = self.next_level_bid(self.opener_bid_one.denomination, '0473')
        elif self._opener_has_bid_two_nt_and_five_four():
            bid = self.nt_bid(3, '0540')
        elif self._five_five_majors_after_3nt():
            bid = self.suit_bid(4, self.heart_suit, '0771')
        elif self._opener_has_bid_same_suit_three_times_and_not_at_game():
            bid = self.bid_to_game(self.opener_bid_one.denomination, '0000')
        elif self._opener_is_strong_and_has_bid_two_nt():
            bid = self.nt_bid(3, '0809')
        elif (self.opener_bid_three.level == 4 and self.opener_bid_three.denomination.is_minor and
              self.suit_length(self.opener_bid_three.denomination) >= 3 and self.hcp >= 11):
            bid = self.next_level_bid(self.opener_bid_three.denomination, '0835')
        else:
            bid = Pass('0533')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _opener_bid_blackwood(self):
        """Return bid after opener has bid Blackwood."""
        if self.openers_last_bid.name == '4NT':
            bid = Blackwood(self.cards, self.board).ace_count
        elif self.openers_last_bid.name == '5NT':
            bid = Blackwood(self.cards, self.board).king_count
        else:
            bid = Pass('0000')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _responder_bid_nt(self):
        """Return bid after responder has bid NT."""
        if self.previous_bid.name == '4NT':
            bid = self._after_my_4nt()
        elif self.previous_bid.name == '4NT':
            bid = Blackwood(self.cards, self.board).select_contract()
        elif self.previous_bid.name == '5NT':
            bid = Blackwood(self.cards, self.board).select_contract()
        else:
            bid = self.nt_bid(1, '0000')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def suit_preference(self):
        """Return suit preference."""
        suit_one_holding = self.suit_length(self.opener_bid_one.denomination)
        suit_two_holding = self.suit_length(self.opener_bid_two.denomination)
        if suit_one_holding >= suit_two_holding:
            suit = self.opener_bid_one.denomination
        else:
            suit = self.opener_bid_two.denomination
        return suit

    def _after_my_4nt(self):
        """Return bid after partner has responded to my 4NT bid."""
        if Blackwood(self.cards, self.board).has_four_aces():
            bid = self.nt_bid(5, '0534')
        elif (self.opener_bid_one.denomination == self.opener_bid_two.denomination and
                self.opener_bid_one.denomination != self.responder_bid_one.denomination and
                self.suit_length(self.opener_bid_one.denomination) >= 2):
            bid = self.suit_bid(6, self.opener_bid_one.denomination, '0512')
        else:
            bid = self.next_level_bid(self.suit_preference(), '0498')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _two_club_opening(self):
        if self._suit_support_after__two_club_opening_opening():
            bid = self.next_level_bid(self.longest_suit, '0530')
        elif self._opener_changes_suit_after__two_club_opening_opening():
            suit = self.suit_preference()
            if self.suit_length(suit) <= 2 and self.nt_level <= 3:
                bid = self.nt_bid(3, '0531')
            else:
                bid = self.next_level_bid(suit, '0532')
        elif self.opener_bid_three.is_nt and self.opener_bid_three.level >= 4:
            bid = self._opener_bid_blackwood()
        elif self.shape[1] >= 5 and self.next_level(self.second_suit) <= 5:
            bid = self.next_level_bid(self.second_suit, '0865')
        else:
            bid = Pass('0886')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _opener_has_rebid_second_suit(self):
        """Return bid if opener has rebid second suit."""
        if self.hcp >= 15:
            bid = self.nt_bid(4, '0000')
        elif (self.partner_bid_one.denomination.is_suit and
                self.partner_bid_one.denomination != self.partner_bid_two.denomination and
                self.partner_last_bid.denomination == self.partner_bid_two.denomination and
                self.suit_length(self.partner_bid_one.denomination) > self.suit_length(self.partner_bid_two.denomination)):
            bid = self.next_level_bid(self.partner_bid_one.denomination, '0815')
        elif self.opener_bid_three.is_game:
            bid = Pass('0361')
        else:
            if (self.shape[0] >= 6 and
                    self.suit_points(self.longest_suit) >= 8 and
                    self.next_level(self.longest_suit) <= 4 and
                    self.longest_suit not in self.opponents_suits):
                suit = self.longest_suit
            elif self.opener_bid_two.denomination == self.responder_bid_two.denomination:
                suit = self.opener_bid_two.denomination
            elif self.suit_length(self.opener_bid_one.denomination) >= 2:
                suit = self.opener_bid_one.denomination
            elif self.suit_length(self.opener_bid_two.denomination) >= 2:
                suit = self.opener_bid_two.denomination
            else:
                suit = self.opener_bid_one.denomination
            bid = self.next_level_bid(suit, '0852')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    # Various boolean functions

    def _opener_bids_three_nt_after_change_of_major(self):
        """Return True if opener bids 3NT after change of suit."""
        result = (self.opener_bid_three.name == '3NT' and
                  self.opener_bid_one.denomination != self.opener_bid_two.denomination and
                  self.opener_bid_one.is_suit_call and
                  self.opener_bid_two.is_suit_call and
                  self.opener_bid_one.denomination.is_major and
                  self.suit_length(self.opener_bid_one.denomination) >= 3)
        return result

    def _opener_bids_three_nt_and_hand_has_six_card_major(self):
        """Return True if opener bids 3NT and hand has 6 card major."""
        result = (self.opener_bid_three.name == '3NT' and
                  self.shape[0] >= 6 and
                  self.longest_suit.is_major and
                  self.longest_suit not in self.opponents_suits and
                  self.next_level(self.longest_suit) <= 4)
        return result

    def _opener_rebids_second_suit_at_three_level(self):
        """Return True if opener rebids second suit at three level."""
        result = (self.opener_bid_three.denomination == self.bid_two.denomination and
                  self.bid_one.denomination != self.bid_two.denomination and
                  self.opener_bid_three.is_suit_call and
                  self.opener_bid_three.level == 3 and
                  self.next_level(self.bid_two.denomination) <= 4 and
                  self.hcp >= 9)
        return result

    def _can_support_openers_second_suit_and_twelve_points(self):
        """Return True if can support openers second suit with opening hand."""
        result = (self.openers_last_bid.level == 3 and
                  (self.openers_last_bid.denomination == self.longest_suit or
                   self.openers_last_bid.denomination == self.second_suit) and
                  self.hcp >= 12)
        return result

    def _opener_jumped_and_can_support(self):
        """Return True if opener jumps and can support."""
        result = (self.is_jump(self.bid_one, self.opener_bid_two) and
                  self.bid_one.denomination == self.opener_bid_three.denomination and
                  self.hcp >= 16 and
                  self.nt_level <= 4)
        return result

    def _opener_bid_minor_at_level_three_twelve_points(self):
        """Return True if opener bids minor at level 3."""
        result = (self.opener_bid_three.level == 3 and
                  self.opener_bid_three.denomination.is_minor and
                  self.hcp >= 12 and
                  self.is_balanced and
                  self.nt_level <= 3)
        return result

    def _opener_rebids_first_suit_two_card_support(self):
        """Return True if opener rebids opening suit and two card support."""
        result = (self.opener_bid_one.denomination == self.opener_bid_three.denomination and
                  self.opener_bid_one.denomination != self.opener_bid_two.denomination and
                  not self.responder_bid_two.is_nt and
                  self.suit_length(self.opener_bid_one.denomination) >= 2 and
                  self.opener_bid_three.level <= 4 and
                  not self.opener_bid_three.is_game and
                  self.hcp >= 8)
        return result

    def _suit_support_after__two_club_opening_opening(self):
        """Return True if after 2C can support."""
        result = (self.opener_bid_one.name == '2C' and
                  self.hcp >= 8 and
                  (self.opener_bid_two.denomination == self.longest_suit or
                   self.opener_bid_three.denomination == self.longest_suit) and
                  self.shape[0] >= 6)
        return result

    def _opener_changes_suit_after__two_club_opening_opening(self):
        """Return True if opener changes suit after 2C opening."""
        result = (self.opener_bid_one.name == '2C' and
                  self.opener_bid_two.denomination != self.opener_bid_three.denomination and
                  not self.openers_last_bid.is_game)
        return result

    def _opener_strong_and_balanced_and_fifteen_points(self):
        """Return True if opener is strong and balanced."""
        result = (self.opener_bid_two.name == '2NT' and
                  self.opener_bid_three.name == '3NT' and
                  self.hcp >= 15)
        return result

    def _two_card_support_for_openers_six_card_suit_twelve_points(self):
        """Return True if two card support for opener's six card suit."""
        result = (self.opener_bid_one.denomination.is_major and
                  self.opener_bid_one.denomination == self.opener_bid_two.denomination and
                  self.opener_bid_two.denomination == self.openers_last_bid.denomination and
                  self.suit_length(self.opener_bid_one.denomination) >= 2 and
                  self.hcp >= 12 and
                  self.next_level(self.opener_bid_one.denomination) <= 4)
        return result

    def _opener_at_level_three_nine_points_and_stoppers(self):
        """Return True if nine points and stoppers."""
        result = (self.openers_last_bid.level == 3 and
                  not(self.opener_bid_one.denomination == self.opener_bid_two.denomination and
                      self.opener_bid_one.denomination == self.opener_bid_three.denomination) and
                  self.hcp >= 11 and
                  self.stoppers_in_unbid_suits and
                  self.nt_level <= 3)
        return result

    def _opener_repeats_second_suit(self):
        """Return True if opener repeats second_suit."""
        result = (self.opener_bid_two.is_suit_call and
                  self.opener_bid_one.denomination != self.opener_bid_two.denomination and
                  self.opener_bid_two.denomination == self.opener_bid_three.denomination
                  and self.nt_level <= 4)
        return result

    def _opener_support_six_card_suit(self):
        """Return True if opener support six card suit."""
        result = (self.opener_bid_three.denomination == self.longest_suit and
                  self.shape[0] >= 6 and
                  self.hcp >= 8 and
                  self.next_level(self.longest_suit) <= 6)
        return result

    def _has_support_for_openers_new_suit_at_level_four(self):
        """Return True if opener shows new suit at 4 level and support."""
        result = (self.opener_bid_three.level == 4 and
                  self.opener_bid_three.denomination != self.opener_bid_one.denomination and
                  self.opener_bid_three.denomination != self.opener_bid_two.denomination and
                  self.suit_length(self.opener_bid_three.denomination) >= 3 and
                  not self.opener_bid_three.is_game)
        return result

    def _support_after_opener_shows_six_four(self):
        """Return True if support for opener who shows six four."""
        result = (self.hcp >= 9 and
                  self.opener_bid_one.denomination == self.opener_bid_two.denomination and
                  self.opener_bid_one.denomination != self.opener_bid_three.denomination and
                  not self.opener_bid_three.is_nt and
                  # self.responder_bid_two.is_nt and
                  not Bid(self.bid_history[-6]).is_pass and
                  self.suit_length(self.opener_bid_one.denomination) >= 2 and
                  not (self.partner_last_bid.denomination == self.opener_bid_one.denomination and
                       self.partner_last_bid.is_game))
        return result

    def _opener_has_bid_two_nt_and_five_four(self):
        """Return True if opener hand bid 2NT and hand is 5/4."""
        result = (self.opener_bid_three.name == '2NT' and
                  self.five_four and
                  self.hcp >= 10 and
                  self.nt_level <= 3)
        return result

    def _has_support_for_openers_six_card_suit(self):
        """Return True if opener has 6 card suit and support."""
        result = (self.opener_bid_one.denomination == self.opener_bid_three.denomination and
                  self.opener_bid_one.denomination != self.opener_bid_two.denomination and
                  self.suit_length(self.opener_bid_one.denomination) >= 2 and
                  self.hcp >= 12 and
                  self.opener_bid_three.level <= 6 and
                  self.next_level(self.opener_bid_one.denomination) <= 5 and
                  not self.opener_bid_three.is_game)
        return result

    def _five_five_majors_after_3nt(self):
        """Return True if _five_five_majors_after_3nt."""
        result = (self.partner_last_bid.name == '3NT' and
                  self.spades >= 5 and self.hearts >= 5 and
                  self.next_level(self.heart_suit) <= 4 and
                  self.heart_suit not in self.opponents_suits)
        return result

    def _opener_has_bid_same_suit_three_times_and_not_at_game(self):
        """Return True if opener has bid suit 3 times and not at game."""
        result = (self.opener_bid_one.denomination == self.opener_bid_two.denomination and
                  self.opener_bid_one.denomination == self.opener_bid_three.denomination and
                  self.suit_length(self.opener_bid_one.denomination) >= 3 and
                  not self.bidding_above_game and
                  self.hcp >= 12)
        return result

    def _opener_is_strong_and_has_bid_two_nt(self):
        """Return True if opener is strong and has bid 2NT."""
        result = (self.opener_bid_three.name == '2NT' and
                  self.hcp >= 8 and
                  self.opener_bid_one.denomination != self.opener_bid_two.denomination)
        return result

    # def xxx(self):
    #     """Return True if xxx."""
    #     result = False
    #     return result
    #
    # def xxx(self):
    #     """Return True if xxx."""
    #     result = False
    #     return result



""" Bid for Game
    Acol AdvancersRebid module
"""

import inspect
from .bidding import Bid, Pass
from .hand import Hand


class AdvancersRebid(Hand):
    """BfG AdvancersRebid class."""
    def __init__(self, hand_cards, board):
        super(AdvancersRebid, self).__init__(hand_cards, board)
        if len(self.bid_history) >= 6:
            self.overcaller_bid_one = Bid(self.bid_history[-6], '')
            self.overcaller_bid_two = Bid(self.bid_history[-2], '')
        else:
            self.overcaller_bid_one = Bid(self.bid_history[-2], '')
            self.overcaller_bid_two = None
        self.bid_one = Bid(self.bid_history[-4], '')
        self.overcaller_repeats_suit = (self.overcaller_bid_one.denomination ==
                                        self.overcaller_bid_two.denomination)
        self.trace = 0

    def suggested_bid(self):
        """Direct control to relevant method and return a Bid object."""
        if self.overcaller_bid_one.is_double:
            bid = self._overcaller_has_doubled()
        else:
            bid = self._overcaller_has_not_doubled()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _overcaller_has_doubled(self):
        """Return bid after overcaller has doubled."""
        if self._has_four_card_support_for_overcallers_second_suit():
            bid = self._double_and_support_second_suit()
        elif self._has_eight_points_and_no_support_for_second_suit():
            bid = self._eight_points_and_no_support_for_second_suit()
        elif self.overcaller_bid_one.is_double and self.overcaller_bid_two.is_double:
            bid = self._two_doubles()
        elif self._three_card_support_for_overcallers_second_suit():
            bid = self.next_level_bid(self.overcaller_bid_two.denomination, '0142')
        elif self._has_eight_points_and_overcaller_passed_suit():
            bid = self.next_level_bid(self.longest_suit, '0647')
        else:
            bid = self._overcaller_has_not_doubled()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _eight_points_and_no_support_for_second_suit(self):
        """Return a bid with 8 points and no support for overcallers second suit."""
        if self._can_bid_at_this_level():
            bid = self._overcaller_has_doubled_suit_bid()
        elif self._has_three_cards_in_overcallers_major():
            bid = self.bid_to_game(self.overcaller_bid_two.denomination, '0025')
        elif self.stoppers_in_bid_suits:
            bid = self.next_nt_bid('0778')
        else:
            bid = self.next_level_bid(self.overcaller_bid_two.denomination, '0025')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _overcaller_has_doubled_suit_bid(self):
        """Return bid after overcaller has doubled and shown suit."""
        if self._has_three_cards_in_overcallers_minor_at_level_four():
            bid = self.next_level_bid(self.overcaller_bid_two.denomination, '0779')
        elif self._can_bid_second_suit_if_five_four():
            bid = self.next_level_bid(self.second_suit, '0780')
        elif self.suit_length(self.overcaller_bid_two.denomination) >= 3:
            if self.hcp >= 10:
                raise_level = 1
            else:
                raise_level = 1
            bid = self.next_level_bid(self.overcaller_bid_two.denomination, '0782', raise_level)
        elif self.longest_suit not in self.opponents_suits:
            bid = self.next_level_bid(self.longest_suit, '0783')
        else:
            bid = Pass('0784')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _overcaller_has_not_doubled(self):
        """Direct control to relevant method and return a Bid object."""
        if self.overcaller_bid_two.level >= 3:
            bid = self._respond_to_three_level()
        elif self.overcaller_bid_one.name in ['1NT', '2NT'] and self.bid_one.name == '2C':
            bid = self._stayman()
        elif self.overcaller_bid_two.name in ['1NT', '2NT']:
            bid = self._respond_to_nt()
        elif self.partner_bid_two.denomination == self.longest_suit and self.shape[0] >= 7:
            bid = self.next_level_bid(self.longest_suit, '0790')
        elif self._overcaller_has_bid_two_suits():
            bid = self._show_preference()
        elif self._can_bid_five_card_suit_after_overcaller_repeats_suit():
            bid = self._overcaller_repeats_suit()
        elif self._has_two_card_support_for_overcallers_six_card_suit():
            bid = self.next_level_bid(self.overcaller_bid_one.denomination, '0785')
        elif self._has_ten_points_and_four_card_support_for_overcaller():
            raise_level = self._support_raise_level(self.overcaller_bid_one.denomination)
            bid = self.next_level_bid(self.overcaller_bid_one.denomination, '0786', raise_level)
        elif self._can_rebid_seven_card_suit():
            bid = self.next_level_bid(self.longest_suit, '0787')
        else:
            bid = Pass('0788')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _overcaller_repeats_suit(self):
        """Return Bid after overcaller has repeated suit with strong hand."""
        if self.suit_length(self.overcaller_bid_one.denomination) >= 2:
            bid = self.next_level_bid(self.overcaller_bid_one.denomination, '0789')
        elif self.longest_suit not in self.opponents_suits:
            bid = self.next_level_bid(self.longest_suit, '0790')
        elif self.hcp >= 12 and self.nt_level <= 3:
            bid = self.nt_bid(3, '0791')
        elif self.nt_level <= 2:
            bid = self.nt_bid(2, '0792')
        else:
            bid = Pass('0793')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _two_doubles(self):
        """Bid after overcaller has doubled twice."""
        opponents_suit = self.openers_agreed_suit()
        if not self.openers_agreed_suit():
            bid = self._respond_to_three_level()
        elif self.suit_length(opponents_suit) >= 4:
            bid = Pass('0794')
        else:
            bid = self._respond_to_three_level()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _double_and_support_second_suit(self):
        """Bid after overcaller has doubled."""
        suit = self.overcaller_bid_two.denomination
        level = self.next_level(self.overcaller_bid_two.denomination)
        if self.hcp >= 10 and level < 4:
            level = min(4, level + 1)
            bid = self.suit_bid(level, suit, '0795')
        elif self.hcp >= 6 and level <= 3:
            level = min(3, level)
            bid = self.suit_bid(level, suit, '0796')
        elif self.hcp >= 8 and level <= 3:
            level = min(4, level)
            bid = self.suit_bid(level, suit, '0797')
        elif self._has_four_card_support_for_overcallers_second_suit_at_level_three():
            bid = self.next_level_bid(self.overcaller_bid_two.denomination, '0798')
        elif self._overcaller_shows_support_after_double():
            bid = self.next_level_bid(suit, '0300')
        else:
            bid = Pass('0799')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _respond_to_three_level(self):
        """Respond after a 3 level bid."""
        if (self.shape[0] >= 7 and self.longest_suit.is_major and self.next_level(self.longest_suit) <= 4 and
                self.longest_suit not in self.opponents_suits):
            bid = self.bid_to_game(self.longest_suit, '0575')
        elif self.overcaller_bid_two.is_game:
            bid = Pass('0800')
        elif self._can_bid_nt_after_overcaller_minor():
            bid = self._respond_to_three_level_minor()
        elif self._can_support_overcaller_second_suit_at_level_three():
            bid = self._respond_to_three_level_two_card_support()
        elif self._can_support_overcaller_second_suit_at_level_four():
            bid = self._respond_to_three_level_with_four_card_support()
        elif self._has_ten_points_and_can_support_overcaller_at_four_level():
            bid = self.next_level_bid(self.overcaller_bid_two.denomination, '0801')
        elif self._can_support_overcaller_after_nt_and_major_bids():
            bid = self.bid_to_game(self.overcaller_bid_two.denomination, '0802')
        elif self._overcaller_has_jumped_after_double():
            bid = self.nt_bid(3, '0803')
        elif self.overcaller_bid_one.is_double and self.overcaller_bid_two.is_suit_call:
            bid = self._double_and_suit()
        elif self.overcaller_bid_two.is_double:
            bid = self._three_level_doubled()
        elif self._has_six_points_and_overcaller_bids_two_suits():
            bid = self.advancer_preference('0889')
        elif self._can_bid_six_card_major_over_overcallers_minor():
            bid = self.next_level_bid(self.longest_suit, '0322')
        elif self._has_five_four_after_overcaller_bids_minor():
            bid = self._respond_to_three_level_five_four()
        else:
            bid = Pass('0805')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _three_level_doubled(self):
        """Return bid after level 3 bid doubled."""
        if self.longest_suit in self.opponents_suits:
            bid = self._holding_opponents_suits()
        else:
            bid = self.next_level_bid(self.longest_suit, '0806')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _holding_opponents_suits(self):
        """Return bid after level 3 bid doubled holding opponents suits."""
        if self.shape[1] <= 3:
            bid = self.next_nt_bid('0807')
        elif self.second_suit not in self.opponents_suits:
            bid = self.next_level_bid(self.second_suit, '0808')
        else:
            bid = Pass('0809')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _double_and_suit(self):
        """Return bid after overcaller has doubled and shown own suit."""
        if self.hcp <= 6 and self.shape[0] <= 6:
            bid = Pass('0810')
        elif self.suit_length(self.overcaller_bid_two.denomination) >= 3:
            bid = self.next_level_bid(self.overcaller_bid_two.denomination, '0811')
        elif self.five_four_or_better and self.second_suit not in self.opponents_suits:
            bid = self.next_level_bid(self.second_suit, '0812')
        elif self._is_unbalanced_and_can_show_suit():
            suit = self.longest_suit
            if suit in self.opponents_suits:
                suit = self.second_suit
            bid = self.next_level_bid(suit, '0813')
        elif self.hcp >= 16:
            bid = self.next_level_bid(self.second_suit, '0814')
        else:
            bid = self.next_nt_bid('0815')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _respond_to_three_level_minor(self):
        """Respond after a 3 level bid with minor suit."""
        if self.overcaller_bid_two.denomination != self.bid_one.denomination:
            holding_one = self.suit_length(self.overcaller_bid_one.denomination)
            holding_two = self.suit_length(self.overcaller_bid_two.denomination)
            if holding_two - holding_one > 1:
                bid = Pass('0816')
            else:
                bid = self.next_level_bid(self.overcaller_bid_one.denomination, '0817')
        else:
            bid = Pass('0818')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _respond_to_three_level_two_card_support(self):
        """Respond after a 3 level bid with 2 card support."""
        if self.next_level(self.overcaller_bid_two) <= self.overcaller_bid_two.game_level:
            bid = self.bid_to_game(self.overcaller_bid_two, '0819')
        else:
            bid = Pass('0820')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _respond_to_three_level_with_four_card_support(self):
        """Respond after a 3 level bid with 4 card support."""
        if self._is_strong_and_four_card_support_for_overcallers_suit():
            bid = self.next_level_bid(self.overcaller_bid_two.denomination, '0821')
        elif self._is_strong_and_no_support_for_overcaller_after_one_nt():
            bid = Pass('0822')
        elif self._has_nine_points_and_overcaller_strong():
            bid = self.next_level_bid(self.overcaller_bid_two.denomination, '0823')
        elif self._has_support_for_overcallers_major():
            bid = self.next_level_bid(self.overcaller_bid_two.denomination, '0171')
        else:
            bid = Pass('0824')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _respond_to_three_level_five_four(self):
        """Respond after a 3 level bid with 5/4 hand"""
        if self.bid_one.denomination == self.second_suit or not self.opener_bid_two.is_pass:
            suit = self.longest_suit
        else:
            suit = self.second_suit
        if self._has_thirteen_points_and_overcaller_support():
            bid = Pass('0825')
        elif self._has_two_card_support_for_overcallers_and_second_suit_minor():
            bid = self.next_level_bid(self.overcaller_bid_one.denomination, '0120')
        elif self._can_bid_selected_suit_at_level_four(suit):
            bid = self.next_level_bid(suit, '0826')
        else:
            bid = Pass('0827')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _stayman(self):
        """Return bid after overcaller responds to Stayman."""
        if self.overcaller_bid_two.is_major:
            if self.hcp <= 8:
                level = 3
            else:
                level = 4
            if (self.suit_length(self.overcaller_bid_two.denomination) >= 4 and
                    self.next_level(self.overcaller_bid_two.denomination) <= level):
                bid = self.suit_bid(level, self.overcaller_bid_two.denomination, '0828')
            else:
                level = 2
                if self.hcp > 10:
                    level = 3
                if level >= self.nt_level:
                    bid = self.nt_bid(level, '0829')
                else:
                    bid = Pass('0830')
        else:
            if self.nt_level <= 3:
                if self.hcp >= 10:
                    bid = self.nt_bid(3, '0238')
                else:
                    bid = self.next_nt_bid('0000')
            else:
                bid = Pass('0000')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _respond_to_nt(self):
        """Bid after overcaller has bid NT."""
        if self.is_semi_balanced or self.shape == [4, 4, 4, 1]:
            bid = self._respond_to_nt_semi_balanced()
        elif self.overcaller_bid_one.is_double and self.overcaller_bid_two.name == '2NT':
            bid = self._suit_after_nt()
        elif self.five_four_or_better and self.second_suit not in self.opponents_suits:
            bid = self.next_level_bid(self.second_suit, '0831')
        else:
            bid = Pass('0832')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _suit_after_nt(self):
        """Return bid after DOUBLE, then NT with distributional hand."""
        suit = self.second_suit
        if suit not in self.opponents_suits:
            bid = self.next_level_bid(suit, '0833')
        else:
            bid = Pass('0834')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _respond_to_nt_semi_balanced(self):
        """Bid after overcaller has bid NT and hand is semi-balanced."""
        if self.overcaller_bid_two.name == '1NT' or self.opener_bid_one.name == '1NT':
            bid = self._respond_to_nt_semi_balanced_after_one_nt()
        else:
            bid = self._respond_to_nt_semi_balanced_after_two_nt()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _respond_to_nt_semi_balanced_after_one_nt(self):
        """Bid after overcaller has bid 1NT and hand is semi-balanced."""
        if 8 <= self.hcp <= 9 and self.nt_level <= 2:
            bid = self.nt_bid(2, '0835')
        elif self.hcp >= 10 and self.nt_level <= 3:
            bid = self.nt_bid(3, '0836')
        else:
            bid = Pass('0837')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _respond_to_nt_semi_balanced_after_two_nt(self):
        """Bid after overcaller has bid 2NT and hand is semi-balanced."""
        if (self.hcp >= 6 and self.nt_level <= 3 and not self.bid_one.is_pass):
            bid = self.nt_bid(3, '0838')
        elif self._has_five_points_and_five_card_major():
            bid = self.next_level_bid(self.longest_suit, '0849')
        else:
            bid = Pass('0839')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _show_preference(self):
        """Show preference after overcaller has bid two suits."""
        suit_one = self.overcaller_bid_one.denomination
        suit_two = self.overcaller_bid_two.denomination
        if self.suit_length(suit_one) >= self.suit_length(suit_two):
            suit = suit_one
        else:
            suit = suit_two
        level = self.next_level(suit)
        if self.suit_length(suit_one) >= 5 and self.shape[0]+self.shape[1] >= 10:
            bid = self.bid_to_game(suit_one, '0670')
        elif (self.hcp >= 4 and level <= 2) or self.hcp >= 10:
            bid = self.next_level_bid(suit, '0840')
        else:
            bid = Pass('0841')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    # Various utility functions

    def _support_raise_level(self, suit):
        """Return the level to support overcaller."""
        if suit.is_major:
            game_level = 4
        else:
            game_level = 5
        if self.hcp <= 10:
            raise_level = 0
        elif self.hcp <= 12:
            raise_level = 1
        else:
            raise_level = 2
        while self.next_level(suit, raise_level) > 4:
            raise_level -= 1
        while self.next_level(suit, raise_level) > game_level:
            raise_level -= 1
        return raise_level

    def _overcaller_minimum_rebid(self):
        """Return True if overcaller has made minimum bid."""
        if (self.overcaller_repeats_suit and
                self.overcaller_bid_two.level - self.overcaller_bid_one.level == 1):
            overcaller_minimum = True
        else:
            overcaller_minimum = False
        return overcaller_minimum

    # Various boolean functions

    def _has_four_card_support_for_overcallers_second_suit(self):
        """Return True if 4 card support for overcallers second bid."""
        result = (self.overcaller_bid_one.is_double and
                  self.overcaller_bid_two.is_suit_call and
                  self.suit_length(self.overcaller_bid_two.denomination) >= 4 and
                  self.overcaller_bid_two.denomination.is_major)
        return result

    def _has_eight_points_and_no_support_for_second_suit(self):
        """Return True if 8 points and little support for overcallers second suit."""
        result = (self.overcaller_bid_one.is_double and
                  self.overcaller_bid_two.is_suit_call and
                  self.overcaller_bid_two.denomination != self.longest_suit and
                  self.hcp >= 8 and
                  not self.overcaller_bid_two.is_game)
        return result

    def _has_eight_points_and_overcaller_passed_suit(self):
        """Return True if 8 points and little support for overcallers second suit."""
        result = (self.overcaller_bid_one.is_double and
                  self.overcaller_bid_two.is_pass and
                  self.next_level(self.longest_suit) <= 2 and
                  self.shape[0] >= 5 and
                  self.hcp >= 8 and
                  not self.overcaller_bid_two.is_game and
                  self.longest_suit not in self.opponents_suits)
        return result

    def _three_card_support_for_overcallers_second_suit(self):
        """Return True if can support overcallers second suit if major."""
        result = (self.overcaller_bid_one.is_double and
                  self.overcaller_bid_two.is_suit_call and
                  self.overcaller_bid_two.is_major and
                  self.suit_length(self.overcaller_bid_two.denomination) >= 3 and
                  self.hcp >= 7)
        return result

    def _can_bid_at_this_level(self):
        """Return True if can bid at this level."""
        result = (self.next_level(self.longest_suit) <= 2 or
                  not self.is_semi_balanced or
                  self.nt_level > 3)
        return result

    def _has_three_cards_in_overcallers_major(self):
        """Return True if 3 card support for overcallers major."""
        result = (self.suit_length(self.overcaller_bid_two.denomination) >= 3 and
                  self.overcaller_bid_two.denomination.is_major)
        return result

    def _has_three_cards_in_overcallers_minor_at_level_four(self):
        """Return True if 3 card support for overcallers minor at 4 level."""
        result = (self.overcaller_bid_two.is_minor and
                  self.overcaller_bid_two.level == 4 and
                  self.suit_length(self.overcaller_bid_two.denomination) >= 3)
        return result

    def _can_bid_second_suit_if_five_four(self):
        """Return True if can bid a 5/4 suit."""
        result = (self.five_four and
                  self.second_suit not in self.opponents_suits)
        return result

    def _overcaller_has_bid_two_suits(self):
        """Return True if overcaller has bid two suits."""
        result = (not self.overcaller_repeats_suit and
                  (self.overcaller_bid_one.is_suit_call and
                   self.overcaller_bid_two.is_suit_call))
        return result

    def _can_bid_five_card_suit_after_overcaller_repeats_suit(self):
        """Return True if own 5 card suit after overcaller_repeats suit."""
        result = (self.overcaller_repeats_suit and
                  self.hcp >= 12 and
                  self.shape[0] >= 5)
        return result

    def _has_two_card_support_for_overcallers_six_card_suit(self):
        """Return True if 2 card support for overcallers 6 card suit."""
        result = (self.overcaller_repeats_suit and
                  self.suit_length(self.overcaller_bid_two.denomination) >= 2 and
                  self.hcp >= 9 and
                  self.suit_length(self.opener_bid_one.denomination) <= 2)
        return result

    def _has_ten_points_and_four_card_support_for_overcaller(self):
        """Return True if 10 points and 4 card support for overcaller."""
        result = (self.hcp >= 10 and
                  self.suit_length(self.overcaller_bid_one.denomination) >= 4 and
                  self.overcaller_bid_one.is_suit_call and
                  not self.bidding_above_game and
                  self.next_level_bid(self.overcaller_bid_one.denomination).level <= 4)
        return result

    def _can_rebid_seven_card_suit(self):
        """Return True if can repeat 7 card suit."""
        result = (self.hcp >= 7 and self.shape[0] >= 7 and
                  self.next_level(self.longest_suit) <= 3 and
                  self.longest_suit not in self.opponents_suits and
                  (self.overcaller_bid_two.is_value_call or
                   self.bid_one.is_value_call) and
                  self.bid_one.denomination != self.overcaller_bid_one.denomination)
        return result

    def _has_four_card_support_for_overcallers_second_suit_at_level_three(self):
        """Return True if 4 card support for overcallers second suit."""
        result = (((self.hcp >= 5 and self.opener_bid_one.name == '1NT') or
                   self.hcp >= 9) and
                  self.next_level(self.overcaller_bid_two.denomination) <= 3 and
                  self.suit_length(self.overcaller_bid_two.denomination) >= 4)
        return result

    def _overcaller_shows_support_after_double(self):
        """Return True if overcaller support after he has doubled."""
        result = (self.overcaller_bid_one.is_double and
                  self.overcaller_bid_two.denomination == self.bid_one.denomination and
                  self.overcaller_bid_two.level == 3 and
                  self.hcp >= 8 and
                  self.suit_length(self.bid_one.denomination) >= 5 and
                  self.next_level(self.bid_one.denomination) <= 4)
        return result

    def _can_bid_nt_after_overcaller_minor(self):
        """Return True if can bid NT over overcaller minor."""
        stoppers = (self.stoppers_in_bid_suits or
                    self.overcaller_bid_one.is_nt or self.overcaller_bid_two.is_nt)
        result = (self.hcp >= 7 and
                  self.overcaller_bid_two.denomination.is_minor and
                  not self._overcaller_minimum_rebid() and
                  stoppers and
                  self.nt_level == 3)
        return result

    def _can_support_overcaller_second_suit_at_level_three(self):
        """Return True if can support overcaller with 2 cards at level 3."""
        result = (self.hcp >= 7 and
                  self.suit_length(self.overcaller_bid_two.denomination) >= 2 and
                  self.next_level(self.overcaller_bid_two.denomination) <= 3)
        return result

    def _can_support_overcaller_second_suit_at_level_four(self):
        """Return True if able to support overcallers second suit at the 4 level."""
        result = (self.suit_length(self.overcaller_bid_two.denomination) >= 4 and
                  self.hcp >= 6 and
                  self.next_level(self.overcaller_bid_two.denomination) <= 4)
        return result

    def _has_ten_points_and_can_support_overcaller_at_four_level(self):
        """Return True if can support overcaller repeated suit at level 4."""
        result = (self.overcaller_repeats_suit and
                  self.suit_length(self.overcaller_bid_two.denomination) >= 3 and
                  self.hcp >= 10 and
                  self.next_level(self.overcaller_bid_two.denomination) <= 4 and
                  not self.opener_bid_one.is_nt and
                  not self.opener_bid_two.is_value_call)
        return result

    def _can_support_overcaller_after_nt_and_major_bids(self):
        """Return True if overcaller has bid suit after nt."""
        result = (self.overcaller_bid_one.is_nt and
                  self.overcaller_bid_two.is_suit_call and
                  self.overcaller_bid_two.denomination.is_major and
                  self.suit_length(self.overcaller_bid_two.denomination) >= 3 and
                  self.next_level(self.overcaller_bid_two.denomination) <= 4)
        return result

    def _overcaller_has_jumped_after_double(self):
        """Return True if overcaller has jumped after a double."""
        result = (self.overcaller_bid_two.is_double and
                  self.stoppers_in_bid_suits and
                  self.hcp >= 10 and
                  not self._overcaller_minimum_rebid() and
                  self.nt_level <= 3)
        return result

    def _has_six_points_and_overcaller_bids_two_suits(self):
        """Return True if 8 points and overcaller changes suits."""
        result = (self.overcaller_bid_one.level == 1 and
                  not self.overcaller_repeats_suit and
                  self.overcaller_bid_one.is_suit_call and
                  self.overcaller_bid_two.is_suit_call and
                  self.hcp >= 6)
        return result

    def _can_bid_six_card_major_over_overcallers_minor(self):
        """Return True if can bid six card major over overcaller's minor."""
        result = (self.overcaller_bid_two.denomination.is_minor and
                  self.hcp >= 10 and
                  self.shape[0] >= 6 and
                  self.longest_suit.is_major and
                  self.next_level(self.longest_suit) <= 3 and
                  self.longest_suit not in self.opponents_suits)
        return result

    def _has_five_four_after_overcaller_bids_minor(self):
        """Return True if 5/4 and overcaller rebids minor."""
        result = (self.overcaller_bid_two.denomination.is_minor and
                  self.hcp >= 6 and
                  self.five_four and
                  not (self.overcaller_repeats_suit and
                       self.overcaller_bid_one.denomination == self.bid_one.denomination))
        return result

    def _is_unbalanced_and_can_show_suit(self):
        """Return True if unbalanced and can show a suit."""
        result = (not self.is_balanced and
                  (self.longest_suit not in self.opponents_suits or
                   self.second_suit not in self.opponents_suits))
        return result

    def _is_strong_and_four_card_support_for_overcallers_suit(self):
        """Return True if strong and 4 card support overcallers suit."""
        result = (((self.hcp >= 10 and self.overcaller_bid_two.denomination.is_major) or
                   self.hcp >= 12) and
                  self.suit_length(self.overcaller_bid_two.denomination) >= 4 and
                  self.overcaller_repeats_suit)
        return result

    def _is_strong_and_no_support_for_overcaller_after_one_nt(self):
        """Return True if strong and no support for overcaller."""
        result = (self.hcp <= 11 and
                  not self.overcaller_bid_one.is_nt and
                  self.suit_points(self.overcaller_bid_two.denomination) <= 2)
        return result

    def _has_nine_points_and_overcaller_strong(self):
        """Return True if overcaller strong and 9 points."""
        result = (self.hcp >= 9 and
                  not self.overcaller_repeats_suit and
                  (self.overcaller_bid_one.level >= 2 or
                   self.overcaller_bid_one.is_nt))
        return result

    def _has_support_for_overcallers_major(self):
        """Return True if support of overcallers major."""
        result = (self.hcp+self.support_points(self.overcaller_bid_two.denomination) >= 10 and
                  self.overcaller_bid_two.denomination.is_major)
        return result

    def _has_thirteen_points_and_overcaller_support(self):
        """Return True if 13 points and overcaller supports suit."""
        result = (self.overcaller_bid_two.denomination == self.bid_one.denomination and
                  self.hcp <= 13 and
                  self.next_level(self.bid_one.denomination) >= 4)
        return result

    def _has_two_card_support_for_overcallers_and_second_suit_minor(self):
        """Return True if overcaller strong and repeats suit."""
        result = (self.is_jump(self.opener_bid_one, self.overcaller_bid_one) and
                  self.overcaller_repeats_suit and
                  self.suit_length(self.overcaller_bid_one.denomination) >= 2 and
                  self.second_suit.is_minor)
        return result

    def _can_bid_selected_suit_at_level_four(self, suit):
        """Return True if can bid suit at level 4."""
        result = ((self.next_level(suit) <= 4 or self.hcp >= 8) and
                  self.second_suit.is_minor and
                  not self.bid_one.is_pass and
                  suit not in self.opponents_suits)
        return result

    def _has_five_points_and_five_card_major(self):
        """Return True if 5 points a nd 5 card major."""
        result = (self.hcp >= 5 and self.shape[0] >= 5 and
                  self.longest_suit.is_major and
                  self.next_level(self.longest_suit) <= 3 and
                  self.longest_suit not in self.opponents_suits)
        return result

    # def xxx(self):
    #     """Return True if xxx."""
    #     result = False
    #     return result

    # def xxx(self):
    #     """Return True if xxx."""
    #     result = False
    #     return result



""" Bid for Game
    Bridge objects
"""

from bridgeobjects import Card, SUITS
from bridgeobjects import Call
from .comment_xref import CommentXref
from .strategy_xref import StrategyXref


class Bid(Call):
    """Return BfG Bid class."""
    def __init__(self, name='', call_id='0000', use_shortage_points=False, *args, **kwargs):
        super(Bid, self).__init__(name, *args, **kwargs)
        self.use_shortage_points = use_shortage_points
        self.rtf_comment = ''
        self.rtf_strategy = ''
        # if call_id not in CommentXref.comments:
        #     call_id = '0000'
        self.call_id = call_id

        # the comments dict contains a gettext reference to the comment/ strategy text.
        self.comment_id = ''
        self.strategy_id = ''
        self.comment_html = ''
        self.strategy_html = ''
        self.comment_rst = ''
        self.strategy_rst = ''

    def __repr__(self):
        """Return a __repr__ display value of object."""
        return f'bid: {self._name}, {self.call_id}'

    def get_comments(self):
        """Get the bid comments from the x_refs."""
        self.comment_id = CommentXref.comments[self.call_id]
        if self.comment_id in StrategyXref.strategies:
            self.strategy_id = StrategyXref.strategies[self.comment_id]
        else:
            print(f'---> no strategy_xref record for {self.comment_id}!')
            quit()

        self.comment_html = CommentXref().comment(self.call_id)
        self.strategy_html = StrategyXref().strategy(self.comment_id)

        self.comment_rst = self._convert_html_to_rst(self.comment_html)
        self.strategy_rst = self._convert_html_to_rst(self.strategy_html)

    @staticmethod
    def _convert_html_to_rst(html):
        """Return and html_string as rst."""
        rst = html.replace('<br>', '\n')
        rst = rst.replace('<p> ', '<p>')
        rst = rst.replace('<p>', '\n\n')
        rst = rst.replace('<span style="color:red">', '**')
        rst = rst.replace('<span style="color:green">', '**')
        rst = rst.replace('</span>', '**')
        return rst


class Pass(Bid):
    """Class definition for PASS bid."""
    def __init__(self, call_id='0000', use_shortage_points=False, *args, **kwargs):
        super(Pass, self).__init__(name='P', call_id=call_id, use_shortage_points=use_shortage_points, *args, **kwargs)
        pass


class Double(Bid):
    """Class definition for DOUBLE bid."""
    def __init__(self, call_id='0000', use_shortage_points=False, *args, **kwargs):
        super(Double, self).__init__(name='D', call_id=call_id, use_shortage_points=use_shortage_points, *args, **kwargs)
        pass


class HandSuit(object):
    """Instantiate BfG HandSuit class."""
    SUITS = {'NT': -1, 'S': 0, 'H': 1, 'D': 2, 'C': 3, -1: 'NT', 0: 'S', 1: 'H', 2: 'D', 3: 'C'}

    def __init__(self, suit=None, hand=None, bid_history=None):
        """Initialise class."""
        self.suit = suit
        self.name = suit.name
        if hand:
            self.cards = hand.cards
            self._card_list = [card for card in hand.cards if card.suit == suit]
        else:
            self.cards = []
            self._card_list = []
        if not bid_history:
            bid_history = []
        self.bid_history = bid_history
        self.honours = 0
        self.touching_honours = 0
        self.honour_points = 0
        self.ace_as_single_honour = False
        self.highest_touching_honour = -1
        if self._card_list:
            self.honours = self._honours()
            self.touching_honours = self._touching_honours()
            self.highest_touching_honour = self._highest_touching_honour()
            self.ace_as_single_honour = self._ace_as_single_honour()
            self.length = len(self._card_list)
            self.honour_points = self._honour_points()
        if bid_history:
            self.partners_bid, self.opponents_bid = self._get_bids()

    def __repr__(self):
        """Return a __repr__ display value of object."""
        return 'HandSuit: %s' % self.name

    def suit_quality(self):
        """
           returns the value of the suit quality
        """
        quality = 0
        if Card('A', self.suit.name) in self._card_list:
            quality += 1
        if Card('K', self.suit.name) in self._card_list:
            quality += 1
        if Card('Q', self.suit.name) in self._card_list:
            quality += 0.5
        if Card('J', self.suit.name) in self._card_list:
            quality += 0.5
        return quality

    def has_control(self, hand):
        """Return True if hand has control of that suit."""
        control = False
        suit = self.suit
        if Card('A', suit.name) in self.cards:
            control = True
        elif (Card('K', suit.name) in self.cards and
              hand.suit_holding[suit] >= 2):
            control = True
        elif (Card('Q', suit.name) in self.cards and
              hand.suit_holding[suit] > 3):
            control = True
        elif (Card('J', suit.name) in self.cards and
              hand.suit_holding[suit] >= 4):
            control = True
        return control

    def _touching_honours(self):
        touching_pairs = ['AK', 'KQ', 'QJ', 'JT']
        touching_honours = 1
        touching_honours_hold = 1
        for index, card in enumerate(self._card_list[1:]):
            if self._card_list[index].rank + card.rank in touching_pairs:
                touching_honours += 1
            else:
                if touching_honours_hold < touching_honours:
                    touching_honours_hold = touching_honours
                touching_honours = 1
        touching_honours = max(touching_honours, touching_honours_hold)
        return touching_honours

    def _highest_touching_honour(self):
        highest_touching_honour = None
        if self.cards and self.touching_honours > 1:
            highest_touching_honour = self.cards[0]
            for card in self.cards:
                if card.value == 13 - highest_touching_honour.value:
                    highest_touching_honour = card
                    break
        return highest_touching_honour

    def _ace_as_single_honour(self):
        result = False
        for card in self._card_list:
            if card.name == 'A':
                result = True
            elif card.name in ['TJQK']:
                result = False
                break
        return result

    def _honours(self):
        """Return the number of honours in hand_suit including tens."""
        honours = 0
        for card in self._card_list:
            if card.value >= 9:
                honours += 1
        return honours

    def _honour_points(self):
        value = 0
        for index, card in enumerate(self._card_list[:5]):
            if card:
                value += 4 - index
        return value

    def _get_bids(self):
        partners_bid, opponents_bid = False, False
        bid_history = list(reversed(self.bid_history[:-3]))
        offset = 0
        contract_suit = self._suit_from_bid(bid_history[0])
        for index, bid in enumerate(bid_history[0::2]):
            suit = self._suit_from_bid(bid)
            if suit == self.suit:
                opponents_bid = True
            if contract_suit == self._suit_from_bid(bid):
                offset = index % 2

        for bid in bid_history[offset*2+1::4]:
            suit = self._suit_from_bid(bid)
            if suit == self.suit:
                partners_bid = True

        return partners_bid, opponents_bid

    @staticmethod
    def _suit_from_bid(bid):
        """Return suit from a bid."""
        suit = None
        if len(bid) == 2:
            suit = SUITS[bid[1]].name
        elif len(bid) == 2:
            suit = -1
        return suit



""" Bid for Game
    Acol AdvancersBid module
"""

import inspect
from .bidding import Bid, Pass, Double

old_file = 'ce/acol_bidding.py'
new_file = 'ource/bidding.py'
from .hand import Hand

# uncomment line around 30 for generating compliance tests. Affects hand 361!!!


class AdvancersBid(Hand):
    """BfG AdvancersBid class."""
    def __init__(self, hand_cards, board):
        super(AdvancersBid, self).__init__(hand_cards, board)
        self.trace = 0

    def suggested_bid(self):
        """Direct control to relevant method and return a Bid object."""
        if self._overcall_bid_at_three_level():
            bid = self._respond_to_three_level()
        elif self.overcaller_bid_one.is_double:
            bid = self._overcaller_has_doubled()
        elif (self.shape[0] >= 9 and
                # comment when not running immersion test
                # self.longest_suit not in self.opponents_suits and
                #
                self.next_level(self.longest_suit) <= 4):
            bid = self.suit_bid(4, self.longest_suit, '0592')
        elif self._has_biddable_eight_card_suit():
            bid = self.bid_to_game(self.longest_suit, '0040')
        elif self.overcaller_bid_one.is_nt:
            bid = self._respond_to_nt()
        elif self._has_three_card_support_for_overcallers_major():
            bid = self._support_for_overcaller()
        elif self._is_weak_no_support_for_overcaller():
            bid = self.nt_bid(1, '0103')
        elif self._has_biddable_suit():
            bid = self._long_suit_bid()
        elif self._can_bid_nt_opposition_has_no_fit():
            bid = self._no_support_stoppers()
        elif self._can_support_overcaller():
            bid = self._support_for_overcaller()
        elif self._has_opening_values_can_bid_nt():
            bid = self._bid_nt()
        elif self.hcp >= 12:
            bid = self._long_suit_bid()
        else:
            bid = Pass('0593')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _respond_to_three_level(self):
        """Respond to 3 level bid."""
        jump_bid = self.is_jump(self.opener_bid_one, self.overcaller_bid_one)
        if self.overcaller_bid_one.is_game:
            bid = Pass('0596')
        elif self._can_bid_six_card_suit_at_three_level():
            bid = self.next_level_bid(self.longest_suit, '0882')
        elif self._has_ten_points_and_stoppers():
            bid = self.nt_bid(3, '0597')
        elif self._can_bid_to_game(self.overcaller_bid_one.denomination):
            bid = self.bid_to_game(self.overcaller_bid_one.denomination, '0598')
        elif self.hcp >= 10 and self.suit_length(self.overcaller_bid_one.denomination) >= 4:
            bid = self.next_level_bid(self.overcaller_bid_one.denomination, '0305')
        elif self.hcp >= 7 and self.opener_bid_one.level >= 2 and not jump_bid:
            bid = Pass('0599')
        elif self._has_eight_points_and_stoppers():
            bid = self.nt_bid(3, '0601')
        else:
            bid = Pass('0602')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _overcaller_has_doubled(self):
        """Bid after partner has doubled."""
        if self._has_eleven_points_balanced_and_responder_has_bid_game():
            bid = Double('0603')
        elif self._is_weak_and_intervening_bid():
            bid = Pass('0604')
        elif self.opener_bid_one.name == '1NT':
            bid = self._overcaller_doubled_one_nt()
        else:
            bid = self._overcaller_has_doubled_weak()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _overcaller_doubled_one_nt(self):
        """Bid after overcaller has doubled opener's 1NT."""
        if 4 <= self.hcp <= 7:
            bid = self._long_suit_bid()
        elif not self.responder_bid_one.is_pass:
            bid = self._overcaller_doubled_one_nt_responder_bid()
        elif self.five_card_suit_or_better and self.hcp <= 4:
            suit = self.longest_suit
            bid = self.next_level_bid(suit, '0605')
        else:
            bid = Pass('0606')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _overcaller_doubled_one_nt_responder_bid(self):
        """Bid after overcaller has doubled opener's 1NT, responder has bid."""
        if self.is_semi_balanced and self.stoppers_in_bid_suits:
            bid = self._semi_balanced_stoppers()
        else:
            bid = self._long_suit_bid()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _semi_balanced_stoppers(self):
        """Return bid if semi-balanced with stoppers in bid suits."""
        if self.hcp <= 9 and self.nt_level <= 2:
            bid = self.nt_bid(2, '0607')
        elif self.nt_level <= 3:
            bid = self.nt_bid(3, '0608')
        else:
            bid = self._long_suit_bid()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _overcaller_has_doubled_weak(self):
        """Bid after partner has doubled with a weak hand."""
        if self._weak_and_responder_has_bid():
            bid = Pass('0609')
        elif self._leave_double():
            bid = Pass('0610')
        elif self.is_balanced and self.stoppers_in_bid_suits:
            bid = self._overcaller_has_doubled_weak_balanced()
        else:
            bid = self._long_suit_bid()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _overcaller_has_doubled_weak_balanced(self):
        """Bid after partner has doubled with a weak balanced hand."""
        major_suit = self._find_four_card_major()
        if self.opponents_at_game:
            bid = Pass('0904')
        elif self.hcp <= 5:
            bid = self._long_suit_bid()
        elif major_suit:
            raise_level = self._find_level_for_major_raise(major_suit)
            bid = self.next_level_bid(major_suit, '0611', raise_level)
        elif self.opener_bid_one.level == 2 and self.nt_level <= 2:
            bid = self.nt_bid(2, '0612')
        elif self._has_ten_points_and_opening_bid_is_weak_three():
            bid = self.nt_bid(3, '0613')
        elif self.hcp >= 12 and self.nt_level <= 3:
            bid = self.nt_bid(3, '0614')
        elif self.hcp >= 10 and self.nt_level <= 2:
            bid = self.nt_bid(2, '0615')
        elif self.hcp >= 6 and self.nt_level <= 1:
            bid = self.nt_bid(1, '0616')
        else:
            bid = self._long_suit_bid()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _support_for_overcaller(self):
        """Returns bid in partner's suit relevant to points."""
        if self._has_thirteen_points_and_five_card_major():
            bid = self._long_suit_bid()
        elif self.overcaller_bid_one.level == 1:
            bid = self._support_for_overcaller_level_one()
        elif self.overcaller_bid_one.level == 2:
            bid = self._support_for_overcaller_level_two()
        elif self.hcp >= 12 and not self.bidding_above_game:
            bid = self.next_level_bid(self.overcaller_bid_one.denomination, '0888')
        elif self._has_strength_and_support_for_overcallers_three_level_bid():
            bid = self.next_level_bid(self.overcaller_bid_one.denomination, '0565')
        else:
            bid = Pass('0617', True)
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _support_for_overcaller_level_one(self):
        """Returns bid in partner's suit after level one bid."""
        suit_to_bid = self.overcaller_bid_one.denomination
        if self._has_ten_points_overcaller_bid_minor():
            bid = self._bid_nt()
        elif self._can_support_overcaller_at_given_level(level=2):
            bid = self.suit_bid(2, suit_to_bid, '0618', True)
        elif self._can_support_overcaller_at_given_level(level=3):
            bid = self.suit_bid(3, suit_to_bid, '0619', True)
        elif self._can_support_overcaller_at_given_level(level=4):
            bid = self.suit_bid(4, suit_to_bid, '0620', True)
        else:
            bid = Pass('0621', True)
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _bid_nt(self):
        """Return a NT bid."""
        if self.hcp >= 13 and self.nt_level <= 3:
            bid = self.nt_bid(3, '0622')
        elif self.hcp >= 12 and self.nt_level <= 2:
            bid = self.nt_bid(2, '0623')
        elif self.nt_level <= 1:
            bid = self.nt_bid(1, '0624')
        else:
            bid = Pass('0625')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _support_for_overcaller_level_two(self):
        """Returns bid in partner's suit after level two bid."""
        suit_to_bid = self.overcaller_bid_one.denomination
        hand_value_points = self._get_hand_value_points(suit_to_bid)
        bid_level = self.next_level(suit_to_bid)
        if self._opening_hand_and_overcaller_bid_minor():
            bid = self.nt_bid(3, '0626')
        elif self._ten_points_and_overcaller_bid_minor():
            bid = self.nt_bid(2, '0627')
        elif hand_value_points in range(9, 14) and bid_level <= 3:
            if self.overcaller_has_jumped:
                bid_level = 4
            else:
                bid_level = 3
            bid = self.suit_bid(bid_level, self.overcaller_bid_one.denomination, '0628', True)
        elif self._good_support_for_overcaller():
            bid = self.suit_bid(4, suit_to_bid, '0630', True)
        elif self.hcp >= 13 and bid_level <= 4:
            bid = self.suit_bid(4, suit_to_bid, '0631', True)
        elif hand_value_points >= 11 and bid_level <= 4:
            bid = self.suit_bid(bid_level, suit_to_bid, '0632', True)
        elif self._is_weak_with_some_support_for_overcaller():
            bid = self.suit_bid(bid_level, suit_to_bid, '0633', True)
        else:
            bid = Pass('0634', True)
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _no_support_stoppers(self):
        """Return bid in NT relevant to points."""
        if self.overcaller_bid_one.level == 1:
            bid = self._no_support_stoppers_level_one()
        elif self.overcaller_bid_one.level == 2:
            bid = self._no_support_stoppers_level_two()
        else:
            bid = Pass('0635')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _no_support_stoppers_level_one(self):
        """Return bid in NT relevant to points at level 1."""
        level = self.nt_level
        if self.hcp in range(8, 12) and level <= 1:
            bid = self.nt_bid(1, '0636')
        elif self.hcp in range(12, 16) and level <= 2:
            bid = self.nt_bid(2, '0637')
        elif self.hcp >= 16 and level <= 3:
            bid = self.nt_bid(3, '0638')
        else:
            bid = Pass('0639')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _no_support_stoppers_level_two(self):
        """Return bid in NT relevant to points at level 2."""
        level = self.nt_level
        if self.hcp in range(9, 14) and level <= 2:
            bid = self.nt_bid(2, '0640')
        elif self.hcp >= 14 and level <= 3:
            bid = self.nt_bid(3, '0641')
        else:
            bid = Pass('0642')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _long_suit_bid(self):
        """Return bid based on own long suit."""
        if self.overcaller_bid_one.is_double:
            bid = self._bid_after_double()
        elif self.suit_length(self.longest_suit) >= 6 and self.hcp >= 11:
            bid = self._long_suit_after_suit()
        elif self._has_eleven_points_and_can_bid_nt():
            nt_level = self.nt_level
            if self._can_bid_five_card_major_below_nt_level():
                bid = self._long_suit_after_suit()
            else:
                if self.hcp >= 17 and self.nt_level == 1:
                    nt_level += 2
                if self.hcp >= 15 and self.nt_level == 2:
                    nt_level += 2
                elif self.hcp >= 13 and self.nt_level == 1:
                    nt_level += 1
                nt_level = min(3, nt_level)
                bid = self.nt_bid(nt_level, '0643')
        elif self._has_five_card_suit_and_seven_points():
            bid = self._long_suit_after_suit()
        elif self._has_seven_card_suit_and_shortage_in_overcallers_suit():
            bid = self.next_level_bid(self.longest_suit, '0644')
        elif self.hcp >= 6 and self.suit_length(self.overcaller_bid_one.denomination) >= 3:
            bid = self.next_level_bid(self.overcaller_bid_one.denomination, '0645')
        else:
            bid = Pass('0646')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _bid_after_double(self):
        """Show long suit after overcaller has doubled."""
        suit = self._select_suit_after_double()
        next_level = self.next_level(suit)
        level, comment = self._bid_after_double_get_level(suit, next_level)
        if self.partner_doubled_game:
            bid = Pass('0647')
        elif self._strong_in_openers_suit():
            bid = self._hold_openers_suit()
        elif self._responder_has_passed_after_one_nt_opening():
            bid = Pass('0035')
        elif self._suit_is_minor_and_can_bid_three_nt(suit, level):
            bid = self.next_nt_bid('0648')
        elif self._suit_is_minor_and_cannot_bid_three_nt(level):
            bid = Double('0649')
        elif self._has_opening_hand_and_can_bid_to_game(suit):
            bid = self.bid_to_game(suit, '0650')
        elif self._has_opening_hand_and_can_bid_to_game(self.longest_suit):
            bid = self.bid_to_game(self.longest_suit, '0650')
        elif self._overcaller_has_doubled_after_weak_two():
            if ((self.hcp >= 10 or
                    self.shape[0] >= 6) and
                    self.next_level(suit) <= 3):
                raise_level = 1
            else:
                raise_level = 0
            bid = self.next_level_bid(suit, '0651', raise_level=raise_level)
        elif self._suit_bid_allowed(level, suit):
            bid = self.suit_bid(level, suit, comment)
        elif self._distributional_strong_hand(suit):
            bid = self.next_level_bid(suit, '0652')
        elif self.last_bid.is_pass and suit not in self.opponents_suits:
            bid = self.next_level_bid(suit, '0653')
        elif self._overcaller_has_doubled_one_nt():
            bid = self.next_level_bid(self.longest_suit, '0907')
        else:
            bid = Pass('0654')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _hold_openers_suit(self):
        """Return bid when holding opener's suit."""
        suit = self._select_suit_after_double()
        if self._is_balanced_and_holds_openers_suit():
            bid = self.next_nt_bid('0655')
        elif self._has_unbid_five_four():
            if self.opener_bid_one.level == 2:
                raise_level = 1
            else:
                raise_level = 0
            bid = self.next_level_bid(suit, '0656', raise_level)
        else:
            bid = Pass('0657')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _long_suit_after_suit(self):
        """Show long suit after overcaller has bid suit."""
        if self._cannot_bid_at_this_level():
            bid = Pass('0665')
        elif self.longest_suit.is_minor and self.is_balanced and self.nt_level <= 2:
            bid = self.nt_bid(2, '0666')
        elif self._is_strong_with_good_six_card_major():
            bid = self.bid_to_game(self.longest_suit, '0667')
        elif self._is_strong_with_good_seven_card_minor():
            bid = self.bid_to_game(self.longest_suit, '0668')
        elif self._has_support_for_overcaller_and_better_points():
            bid = self.next_level_bid(self.overcaller_bid_one.denomination, '0747')
        elif self._can_bid_own_suit_after_suit():
            if self.hcp >= 15:
                bid = self.next_level_bid(self.longest_suit, '0112', raise_level=1)
            else:
                bid = self.next_level_bid(self.longest_suit, '0664')
        elif self._is_balanced_after_overcallers_minor():
            bid = self.next_nt_bid('0471')
        elif self.hcp >= 11 and self.suit_length(self.overcaller_bid_one.denomination) >= 3:
            bid = self.next_level_bid(self.overcaller_bid_one.denomination, '0413')
        elif (self.overcaller_bid_one.level >= 2 and
              self.hcp >= 7 and
              self.shape[0] >= 7 and
              self.longest_suit not in self.opponents_suits):
            bid = self.next_level_bid(self.longest_suit, '0472')
        else:
            bid = Pass('0669')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _respond_to_nt(self):
        """Respond to a NT overcall."""
        if self.is_semi_balanced:
            bid = self._respond_to_nt_semi_balanced()
        elif self._has_ten_points_and_can_bid_six_card_major():
            bid = self.bid_to_game(self.longest_suit, '0731')
        elif self._has_six_points_and_can_bid_five_card_major():
            bid = self.next_level_bid(self.five_card_major_suit, '0670')
        else:
            bid = self._respond_to_nt_distributional()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _respond_to_nt_semi_balanced(self):
        """Respond to a NT overcall is semi_balanced."""
        if self.overcaller_bid_one.name == '1NT':
            bid = self._respond_to_one_nt_semi_balanced()
        elif self.overcaller_bid_one.name == '2NT':
            bid = self._respond_to_two_nt_semi_balanced()
        else:
            bid = self._respond_to_two_nt_semi_balanced_weak()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _respond_to_one_nt_semi_balanced(self):
        """Respond to 1NT overcall is semi_balanced."""
        if self._can_bid_stayman_after_overcallers_one_nt():
            bid = self.club_bid(2, '0671')
        elif self._can_bid_stayman_after_overcallers_two_nt():
            bid = self.club_bid(3, '0672')
        elif 8 <= self.hcp <= 9 and self.nt_level <= 2:
            bid = self.nt_bid(2, '0673')
        elif self._has_six_card_major_and_can_bid_game():
            bid = self.suit_bid(4, self.longest_suit, '0067')
        elif self._has_nine_points_and_four_card_major():
            bid = self.next_level_bid(self.longest_suit, '0125', raise_level=1)
        elif self.hcp >= 10 and self.nt_level <= 3:
            bid = self.nt_bid(3, '0674')
        else:
            bid = Pass('0675')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _respond_to_two_nt_semi_balanced(self):
        """Respond to 2NT overcall is semi_balanced."""
        if self._has_eight_points_and_five_card_major():
            bid = self.next_level_bid(self.five_card_major_suit, '0676')
        elif self._overcaller_bid_nt_and_four_card_major():
            bid = self.suit_bid(3, self.club_suit, '0928')
        elif self.hcp >= 8 and self.nt_level <= 3:
            bid = self.nt_bid(3, '0677')
        else:
            bid = self._respond_to_two_nt_semi_balanced_weak()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _respond_to_two_nt_semi_balanced_weak(self):
        """Respond to 2NT overcall is semi_balanced with weak hand."""
        if self.hcp >= 4 and self.nt_level <= 3 and self.opener_bid_one.level == 1:
            bid = self.nt_bid(3, '0678')
        else:
            bid = Pass('0679')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _respond_to_nt_distributional(self):
        """Respond to a NT overcall if distributional."""
        if (self.overcaller_bid_one.name == '2NT' and
                self.opener_bid_one.level == 1):
            bid = self._respond_to_nt_distributional_overcaller_jumped()
        elif self.hcp >= 4:
            bid = self._respond_to_nt_after_weak_2()
        else:
            bid = Pass('0680')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _respond_to_nt_after_weak_2(self):
        """Return bid after opener bid weak 2"""
        suit = self._select_suit()
        bid = self.next_level_bid(suit)
        if bid.name == '2C':
            bid = self._respond_to_nt_distributional_with_clubs()
        elif (self.suit_length(suit) >= 5 and
                suit not in self.opponents_suits and
                self.next_level(suit) < self.bid_to_game(suit).level and
                suit != self.club_suit):
            bid = self.next_level_bid(suit, '0681')
        elif self._is_five_four_or_five_five():
            suit = self.longest_suit
            if suit in self.opponents_suits:
                suit = self.second_suit
            if self.next_level(suit) < self.bid_to_game(suit).level:
                bid = self.next_level_bid(suit, '0682')
            else:
                bid = Pass('0000')
        elif self.opener_bid_one.level == 2 and self.overcaller_bid_one.name == '2NT':
            bid = self._nt_after_weak_opening()
        else:
            bid = Pass('0683')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _nt_after_weak_opening(self):
        """Return bid after overcaller bids NT after weak 2 opening."""
        if self.hcp <= 6:
            bid = Pass('0684')
        elif self.four_card_major_suit and self.next_level(self.club_suit) <= 3:
            bid = self.next_level_bid(self.club_suit, '0685')
        elif self.nt_level <= 3:
            bid = self.nt_bid(3, '0686')
        else:
            bid = Pass('0687')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _respond_to_nt_distributional_with_clubs(self):
        """Respond to a NT overcall if best suit is clubs."""
        if self.hcp >= 8:
            bid = self.nt_bid(2, '0688')
        else:
            bid = Pass('0689')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _respond_to_nt_distributional_overcaller_jumped(self):
        """Respond to a jump NT overcall is distributional."""
        if self._can_bid_five_card_major_at_level_three():
            bid = self.suit_bid(3, self.longest_suit, '0690')
        else:
            bid = Pass('0691')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    # Various utility functions

    def _fourth_suit_bid(self, suit):
        """Return True if suit is fourth suit bid."""
        value = False
        suits = [0, 0, 0, 0]
        suits[suit.rank] = 1
        for bid in self.bid_history:
            if Bid(bid).is_suit_call:
                bid_suit = Bid(bid).denomination.rank
                suits[bid_suit] = 1
        if sum(suits) == 4:
            value = True
        return value

    def _select_suit(self, force_major=False):
        """Return best suit to bid."""
        suit_one, suit_two = self._top_two_unbid_suits()
        suit_one_holding = self.suit_length(suit_one)
        suit_two_holding = self.suit_length(suit_two)
        if (suit_one_holding == suit_two_holding or
                (suit_one_holding == 5 and suit_two_holding == 4 and
                 suit_one.is_minor and suit_two.is_major)):
            if suit_two.is_major:
                suit = self.cheaper_suit(suit_one, suit_two)
            elif suit_one.is_major:
                suit = suit_one
            elif (suit_one.is_major and
                  suit_one_holding == suit_two_holding):
                suit = suit_one
                if not force_major and suit_one_holding == suit_two_holding:
                    suit = self.cheaper_suit(suit_one, suit_two)
            else:
                suit = self.cheaper_suit(suit_one, suit_two)
        else:
            suit = suit_one
        if suit == self.opener_bid_one.denomination or suit == self.responder_bid_one.denomination:
            if suit_two_holding >= 0:
                suit = suit_two
            else:
                suit = self.no_trumps
        return suit

    def _top_two_unbid_suits(self):
        """Return two longest unbid suits."""
        suit_index = 0
        suit_offset = 1
        found = False
        # Loop through suits to find two longest unbid suits.
        suit_one, suit_two = None, None
        while not found:
            suit_one = self.ordered_holding[suit_index][1]
            suit_two = self.ordered_holding[suit_index+suit_offset][1]
            if (suit_one == self.opener_bid_one.denomination or
                    suit_one == self.responder_bid_one.denomination):
                suit_index += 1
            elif (suit_two == self.opener_bid_one.denomination or
                  suit_two == self.responder_bid_one.denomination):
                suit_offset += 1
            else:
                found = True
        return suit_one, suit_two

    def _can_bid_to_game(self, suit):
        """Return True if conditions are met for raise to game."""
        value = False
        sufficient_points = (self.hcp >= 8 or
                             (self.hcp >= 7 and self.shape[3] <= 1))
        if (self.suit_length(suit) >= 2 and
                sufficient_points and
                not self.is_insufficient_bid(self.bid_to_game(suit)) and
                (self.opener_bid_one.level == 1 or
                 suit.is_major)):
            value = True
        return value

    def _find_level_for_major_raise(self, major_suit):
        """Return the appropriate raise level for major suit  support."""
        if self.next_level(major_suit) >= 4:
            raise_level = 0
        elif self.hcp >= 13:
            raise_level = 2
        elif self.hcp >= 10:
            raise_level = 1
        else:
            raise_level = 0
        return raise_level

    def _find_four_card_major(self):
        """Return an unbid four card major."""
        if self.spades >= 4 and self.spade_suit not in self.opponents_suits:
            suit = self.spade_suit
        elif (self.hearts >= 4 and
              self.heart_suit not in self.opponents_suits):
            suit = self.heart_suit
        else:
            suit = None
        return suit

    def _leave_double(self):
        value = False
        suit = self._select_suit()
        openers_suit = Bid(self.bid_history[-3]).denomination
        if (suit.is_minor and
                self.suit_length(openers_suit) >= 4 and
                self.hcp >= 10 and
                Bid(self.bid_history[-1]).is_pass):
            value = True
        return value

    def _get_hand_value_points(self, suit):
        """Return hand value points."""
        suit_cards = self.suit_length(suit)
        hand_value_points = (self.hcp +
                             self.support_shape_points(suit) +
                             suit_cards - 3)
        if self.shape[3] == 0:
            hand_value_points += 1
        return hand_value_points

    def _select_suit_after_double(self):
        five_four = (self.five_four and
                     (self.longest_suit in self.opponents_suits or
                      self.second_suit in self.opponents_suits))
        if self.five_five:
            if (self.longest_suit.rank < self.second_suit.rank and
                    self.second_suit not in self.opponents_suits):
                suit = self.second_suit
            elif self.longest_suit not in self.opponents_suits:
                suit = self.longest_suit
            else:
                suit = self.second_suit
        elif five_four:
            if self.longest_suit not in self.opponents_suits:
                suit = self.longest_suit
            else:
                suit = self.second_suit
        else:
            suit = self._select_suit(force_major=True)
        return suit

    def _suit_bid_allowed(self, level, suit):
        """Return True if a suit bid is allowed."""
        double_forcing = Bid(self.bid_history[-1]).is_pass
        sufficient_points = (self.hcp >= 9 or level <= 2)
        level_low_enough = (self.next_level(suit) <= 4 and level >= self.next_level(suit))

        if not double_forcing and self.suit_length(suit) < 4:
            return False
        elif double_forcing and suit not in self.opponents_suits:
            return True
        elif (sufficient_points and
                level_low_enough and
                suit not in self.opponents_suits):
            return True
        return False

    def _bid_after_double_get_level(self, suit, level):
        """Return level and comment with a long suit after a double."""
        comment = '0658'
        if self.opener_bid_one.level == 1:
            if self.responder_bid_one.is_value_call:
                if self.hcp >= 12:
                    level += 1
                comment = '0659'
            else:
                comment = '0660'
                if (self.hcp >= 13 and suit.is_major and
                        self.suit_length(suit) >= 5):
                    level = 4
                    comment = '0661'
                elif self.hcp >= 12:
                    level += 2
                elif self.hcp >= 9 and level == 1:
                    level += 1
        elif self.opener_bid_one.level == 2:
            comment = '0662'
            level = self.next_level(suit)
            if self.hcp >= 8:
                level = max(3, level)
            else:
                level = max(2, level)
        elif self.opener_bid_one.level == 3:
            comment = '0663'
            level = self.next_level(suit)
            if self.hcp >= 8:
                level = max(4, level)
            else:
                level = max(3, level)
        return level, comment

    # Various boolean functions

    def _overcall_bid_at_three_level(self):
        """Return True if overcaller has bid at level 3."""
        result = (self.overcaller_bid_one.level >= 3 and
                  self.opener_bid_one.level != 3)
        return result

    def _has_biddable_eight_card_suit(self):
        """Return True if with biddable 8 card suit."""
        result = (self.shape[0] == 8 and
                  self.next_level(self.longest_suit) <= self.longest_suit.game_level and
                  self.longest_suit not in self.opponents_suits)
        return result

    def _has_three_card_support_for_overcallers_major(self):
        """Return True if 3 card support for overcaller's major."""
        result = (self.overcaller_bid_one.is_major and
                  self.suit_length(self.overcaller_bid_one.denomination) >= 3)
        return result

    def _is_weak_no_support_for_overcaller(self):
        """Return True if weak and no support for overcaller."""
        result = (self.bid_history[-1] == 'P' and
                  8 <= self.hcp <= 9 and
                  self.suit_length(self.overcaller_bid_one.denomination) <= 2 and
                  self.shape[0] <= 5 and
                  self.nt_level == 1 and
                  self.stoppers_in_bid_suits)
        return result

    def _has_biddable_suit(self):
        """Return True if hand has a biddable suit."""
        biddable_six_card_suit = (self.shape[0] >= 6 and
                                  self.suit_length(self.overcaller_bid_one.denomination) <= 2)
        can_compete_after_nt = (self.opener_bid_one.name == '1NT' and
                                self.overcaller_bid_one.is_suit_call and
                                self.overcaller_bid_one.level == 2 and
                                self.hcp >= 10)
        if (self.suit_length(self.longest_suit) == self.suit_length(self.second_suit) and
                self.second_suit == self.overcaller_bid_one.denomination):
            my_suit = self.second_suit
        else:
            my_suit = self.longest_suit
        has_own_five_card_suit = (self.suit_length(self.overcaller_bid_one.denomination) <= 1 and self.shape[0] >= 5)
        has_own_five_card_major = (self.hcp >= 8 and self.shape[0] >= 5 and my_suit.is_major)
        can_bid_own_suit = (self.hcp >= 8 and
                            self.shape[0] >= 5 and
                            self.responder_bid_one.is_value_call and
                            self.overcaller_bid_one.is_minor and
                            my_suit.is_major)
        result = (biddable_six_card_suit or
                  can_bid_own_suit or
                  can_compete_after_nt or
                  has_own_five_card_major or
                  has_own_five_card_suit)
        return result

    def _can_bid_nt_opposition_has_no_fit(self):
        """Return True if opposition weak or no fit can bid nt."""
        result = (self.stoppers_in_bid_suits and
                  (not self.responder_bid_one.is_pass or
                   self.hcp >= 16) and
                  (self.opener_bid_one.denomination != self.responder_bid_one.denomination or
                   self.responder_bid_one.level <= 2))
        return result

    def _can_support_overcaller(self):
        """Return True if can support overcaller."""
        if self.overcaller_has_jumped:
            support_for_overcaller = 2
        else:
            support_for_overcaller = 3
        result = self.suit_length(self.overcaller_bid_one.denomination) >= support_for_overcaller
        return result

    def _has_opening_values_can_bid_nt(self):
        """Return True if opening values and can bid nt."""
        result = (self.is_balanced and
                  self.hcp >= 11 and
                  not self.responder_bid_one.is_value_call and
                  self.stoppers_in_bid_suits)
        return result

    def _can_bid_six_card_suit_at_three_level(self):
        """Return True if can bid 6 card suit at level 3."""
        result = (self.shape[0] >= 6 and
                  self.hcp >= 10 and
                  self.next_level(self.longest_suit) <= 3 and
                  self.longest_suit not in self.opponents_suits)
        return result

    def _has_ten_points_and_stoppers(self):
        """Return True if 10 points and can bid 3NT."""
        result = (self.hcp >= 10 and
                  self.overcaller_bid_one.denomination.is_minor and
                  self.stoppers_in_bid_suits and
                  self.nt_level == 3)
        return result

    def _has_eight_points_and_stoppers(self):
        """Return True if eight points and stoppers."""
        result = (self.hcp >= 8 and
                  self.stoppers_in_unbid_suits() and
                  self.nt_level == 3)
        return result

    def _has_eleven_points_balanced_and_responder_has_bid_game(self):
        """Return True if 11 points and responder at game."""
        result = (self.hcp >= 11 and
                  self.shape[0] == 4 and
                  self.responder_bid_one.is_game)
        return result

    def _is_weak_and_intervening_bid(self):
        """Return True if weak after overcaller has doubled."""
        double_index = self.bid_history.index('D')
        next_bid = Bid(self.bid_history[double_index+1])
        result = (next_bid.is_suit_call and
                  self.hcp <= 5 and
                  self.shape[0] <= 5)
        return result

    def _weak_and_responder_has_bid(self):
        """Return True if weak and responder has bid."""
        result = (self.hcp <= 5 and
                  not self.responder_bid_one.is_pass and
                  self.responder_bid_one.level <= 2)
        return result

    def _has_ten_points_and_opening_bid_is_weak_three(self):
        """Return True if 10+ points and opener has bid at 3 level."""
        result = (self.opener_bid_one.level == 3 and
                  self.nt_level <= 3 and
                  self.hcp >= 10)
        return result

    def _has_thirteen_points_and_five_card_major(self):
        """Return True if 13 points and five card major."""
        result = (self.overcaller_bid_one.denomination.is_minor and
                  self.five_card_major_or_better and
                  self.hcp >= 13)
        return result

    def _has_ten_points_overcaller_bid_minor(self):
        """Return True if 10 points and overcaller bid minor."""
        result = (self.overcaller_bid_one.is_minor and
                  self.is_semi_balanced and
                  self.stoppers_in_bid_suits and
                  self.hcp >= 10 and
                  not Bid(self.bid_history[-1]).is_value_call)
        return result

    def _can_support_overcaller_at_given_level(self, level):
        """Return True if can support overcaller at level 2."""
        hand_value_points = self._get_hand_value_points(self.overcaller_bid_one.denomination)
        if level == 2:
            points_range = range(8, 13)
        elif level == 3:
            points_range = range(12, 16)
        elif level == 4:
            points_range = range(16, 40)
        else:
            points_range = range(0, 0)
        result = (hand_value_points in points_range and
                  self.next_level(self.overcaller_bid_one.denomination) <= level)
        return result

    def _opening_hand_and_overcaller_bid_minor(self):
        """Return True if opening hand and overcaller bids minor."""
        result = (self.hcp >= 13 and
                  self.overcaller_bid_one.denomination.is_minor and
                  self.stoppers_in_bid_suits and
                  self.nt_level <= 3)
        return result

    def _ten_points_and_overcaller_bid_minor(self):
        """Return True if 10+ points and overcaller bids minor."""
        result = (self.hcp >= 10 and
                  self.overcaller_bid_one.denomination.is_minor and
                  self.stoppers_in_bid_suits and
                  self.is_semi_balanced and
                  self.nt_level <= 2)
        return result

    def _good_support_for_overcaller(self):
        """Return True if good support for overcaller."""
        hand_value_points = self._get_hand_value_points(self.overcaller_bid_one.denomination)
        result = (self.overcaller_bid_one.level <= 4 and
                  self.next_level(self.overcaller_bid_one.denomination) <= 4 and
                  (self.hcp >= 16 or
                   (hand_value_points >= 14 and
                    self.suit_length(self.overcaller_bid_one.denomination) >= 5)))
        return result

    def _is_weak_with_some_support_for_overcaller(self):
        """Return True if weak with some support for overcaller."""
        result = (self.suit_length(self.overcaller_bid_one.denomination) >= 4 and
                  self.suit_points(self.overcaller_bid_one.denomination) >= 1 and
                  self.hcp >= 5 and
                  self.opener_bid_one.level == 2)
        return result

    def _has_eleven_points_and_can_bid_nt(self):
        """Return True if 11 points and can bid nt."""
        result = (self.stoppers_in_bid_suits and
                  self.hcp >= 11 and
                  self.nt_level <= 3)
        return result

    def _can_bid_five_card_major_below_nt_level(self):
        """Return True if can bid 5 card major."""
        result = (self.next_level(self.longest_suit) <= self.nt_level and
                  self.shape[0] >= 5 and
                  self.longest_suit.is_major)
        return result

    def _has_five_card_suit_and_seven_points(self):
        """Return True if 7 points and 5 card suit."""
        suit = self._select_suit()
        result = (suit.is_suit and
                  self.suit_length(suit) >= 5 and
                  self.hcp >= 7 and
                  not self.opener_bid_two.is_nt)
        return result

    def _has_seven_card_suit_and_shortage_in_overcallers_suit(self):
        """Return True if 7 card suit and shortage in overcallers suit."""
        result = (self.suit_length(self.overcaller_bid_one.denomination) <= 1 and
                  self.shape[0] >= 7 and
                  self.hcp >= 3 and
                  self.next_level(self.longest_suit) <= 2 and
                  self.longest_suit not in self.opponents_suits)
        return result

    def _strong_in_openers_suit(self):
        """Return True if strong in openers suit."""
        result = (self.opener_bid_one.is_suit_call and
                  ((self.hcp >= 8 and
                    self.suit_length(self.opener_bid_one.denomination) >= 4) or
                   (self.hcp >= 6 and
                    self.suit_length(self.opener_bid_one.denomination) >= 5)))
        return result

    def _responder_has_passed_after_one_nt_opening(self):
        """Return True if 7 points and responder passes after 1NT."""
        result = (self.opener_bid_one.is_nt and
                  self.hcp >= 7 and
                  self.responder_bid_one.is_pass)
        return result

    def _suit_is_minor_and_can_bid_three_nt(self, suit, level):
        """Return True if suit is minor and level over 3 NT."""
        result = (10 <= self.hcp <= 12 and
                  level >= 4 and
                  suit.is_minor and
                  self.stoppers_in_bid_suits)
        return result

    def _suit_is_minor_and_cannot_bid_three_nt(self, level):
        """Return True if suit is minor and level over 3 NT."""
        result = (self.hcp >= 8 and
                  level >= 4 and
                  self.responder_bid_one.denomination.is_minor and
                  not self.overcaller_bid_one.is_double)
        return result

    def _has_opening_hand_and_can_bid_to_game(self, suit):
        """Return True if opening hand and can bid to game in suit."""
        double_level = self.double_level()
        game_level = self.game_level(suit)
        next_level = self.next_level(suit)
        result = (self.hcp >= 13 and
                  double_level == 3 and
                  next_level <= game_level)
        return result

    def _overcaller_has_doubled_after_weak_two(self):
        """Return True if overcaller has doubled after weak two opening."""
        result = (self.hcp >= 6 and
                  self.opener_bid_one.level == 2 and
                  self.longest_suit not in self.opponents_suits)
        return result

    def _distributional_strong_hand(self, suit):
        """Return True if distributional hand."""
        result = (self.hcp + self.distribution_points >= 12 and
                  self.next_level(suit) <= 4 and
                  suit not in self.opponents_suits)
        return result

    def _overcaller_has_doubled_one_nt(self):
        """Return True if overcaller has doubled 1NT."""
        result = (self.opener_bid_one.name == '1NT' and
                  self.overcaller_bid_one.is_double and
                  self.hcp >= 6 and
                  self.longest_suit not in self.opponents_suits)
        return result

    def _is_balanced_and_holds_openers_suit(self):
        """Return True if balanced and holds openers suit."""
        result = (self.hcp >= 8 and
                  self.is_balanced and
                  self.stoppers_in_bid_suits)
        return result

    def _has_unbid_five_four(self):
        """Return True if unbid 5/4 suits."""
        result = (self.hcp >= 8 and
                  self.five_four and
                  (self.longest_suit not in self.opponents_suits or
                   self.second_suit not in self.opponents_suits))
        return result

    def _cannot_bid_at_this_level(self):
        """Return True if cannot make a sensible bid at this level."""
        level = self.next_level(self.longest_suit)
        jump = self.is_jump(self.opener_bid_one, self.overcaller_bid_one)
        cannot_bid_at_level_three = (level >= 3 and self.hcp <= 11 and self.shape[0] <= 4 and not jump)
        cannot_bid_at_level_two = (level >= 2 and self.hcp <= 9 and self.shape[0] <= 5 and not jump)
        result = (level >= 7 or
                  cannot_bid_at_level_three or
                  cannot_bid_at_level_two or
                  self.hcp < 6)
        return result

    def _is_strong_with_good_six_card_major(self):
        """Return True if strong with good six card major."""
        result = (self.shape[0] >= 6 and
                  self.hcp >= 15 and
                  self.suit_points(self.longest_suit) >= 8 and
                  self.longest_suit.is_major and
                  self.longest_suit not in self.opponents_suits)
        return result

    def _is_strong_with_good_seven_card_minor(self):
        """Return True if strong with good 7 card minor."""
        result = (self.shape[0] >= 7 and
                  self.hcp >= 15 and
                  self.suit_points(self.longest_suit) >= 8 and
                  self.longest_suit.is_minor and
                  self.longest_suit not in self.opponents_suits)
        return result

    def _can_bid_own_suit_after_suit(self):
        """Return True if can bid own suit."""
        result = ((self.next_level(self.longest_suit) <= 2 or
                   self.hcp >= 10 or
                   self.longest_suit.is_major) and
                  (self.hcp >= 10 or not self._fourth_suit_bid(self.longest_suit)) and
                  self.longest_suit not in self.opponents_suits)
        return result

    def _has_six_points_and_can_bid_five_card_major(self):
        """Return True if 6+ points and can bid 5 card major."""
        result = (self.five_card_major_suit and
                  self.five_card_major_suit not in self.opponents_suits and
                  self.hcp >= 6 and
                  self.next_level(self.longest_suit) <= 4)
        return result

    def _has_ten_points_and_can_bid_six_card_major(self):
        """Return True if 10+ points and can bid 6 card major."""
        result = (self.shape[0] >= 6 and
                  self.longest_suit.is_major and
                  self.longest_suit not in self.opponents_suits and
                  self.hcp >= 10 and
                  self.next_level(self.longest_suit) <= 4)
        return result

    def _can_bid_stayman_after_overcallers_one_nt(self):
        """Return True if 4 card major over overcallers 1NT."""
        result = (self.overcaller_bid_one.name == '1NT' and
                  self.four_card_major_suit and
                  self.suit_length(self.four_card_major_suit) == 4 and
                  self.hcp >= 8 and
                  self.next_level(self.club_suit) <= 2)
        return result

    def _can_bid_stayman_after_overcallers_two_nt(self):
        """Return True if 4 card major over overcallers 2NT."""
        result = (self.overcaller_bid_one.name == '2NT' and
                  self.four_card_major_suit and
                  self.hcp >= 6)
        return result

    def _has_six_card_major_and_can_bid_game(self):
        """Return True if six_card_major."""
        result = (self.shape[0] >= 6 and
                  self.longest_suit.is_major and
                  self.next_level(self.longest_suit) <= 4 and
                  self.longest_suit not in self.opponents_suits)
        return result

    def _has_nine_points_and_four_card_major(self):
        """Return True if has 9 points and 4 card major."""
        result = (self.four_card_major_or_better and
                  self.hcp >= 9 and
                  self.longest_suit not in self.opponents_suits)
        return result

    def _has_eight_points_and_five_card_major(self):
        """Return True if 6 points and a five card major."""
        result = (self.five_card_major_suit and
                  self.five_card_major_suit not in self.opponents_suits and
                  self.hcp >= 8)
        return result

    def _is_five_four_or_five_five(self):
        """Return True if 5/4 or 5/5."""
        result = ((self.five_four or
                   self.five_five) and
                  (self.longest_suit not in self.opponents_suits or
                   self.second_suit not in self.opponents_suits))
        return result

    def _can_bid_five_card_major_at_level_three(self):
        """Return True if can bid 5 card major at level 3."""
        result = (self.longest_suit.is_major and
                  self.longest_suit not in self.opponents_suits and
                  self.suit_length(self.longest_suit) >= 5 and
                  self.next_level(self.longest_suit) == 3)
        return result

    def _is_balanced_after_overcallers_minor(self):
        """Return True if balanced and stoppers in bid suit."""
        result = (self.hcp >= 11 and
                  self.is_balanced and
                  self.overcaller_bid_one.denomination.is_minor and
                  self.stoppers_in_bid_suits)
        return result

    def _has_strength_and_support_for_overcallers_three_level_bid(self):
        """Return True if 10 points and supporter for overcaller."""
        result = (self.overcaller_bid_one.level == 3 and
                  self.overcaller_bid_one.denomination.is_major and
                  self.hcp >= 10 and
                  self.suit_length(self.overcaller_bid_one.denomination) >= 3 and
                  self.next_level(self.overcaller_bid_one.denomination) <= 4)
        return result

    def _has_support_for_overcaller_and_better_points(self):
        """Return True if support and better points in overcaller's suit."""
        result = (self.suit_length(self.overcaller_bid_one.denomination) >= 4 and
                  self.suit_points(self.overcaller_bid_one.denomination) >= 5 and
                  self.suit_points(self.longest_suit) <= 3)
        return result

    def _overcaller_bid_nt_and_four_card_major(self):
        """Return True if overcall has bid no trumps and a 4 card major."""
        result = (self.hcp >= 8 and
                  self.longest_suit.is_major and
                  self.next_level(self.club_suit) <= 3)
        return result

    # def xxx(self):
    #     """Return True if xxx."""
    #     result = False
    #     return result

    # def xxx(self):
    #     """Return True if xxx."""
    #     result = False
    #     return result


""" Bid for Game
    Player class
"""
from termcolor import cprint

from bridgeobjects import ROLES
from .acol_bidding import AcolBid

MODULE_COLOUR = 'blue'

class Player(object):
    """Define BfG Player class."""
    NUMBER_OF_PLAYERS = 4

    def __init__(self, parent=None, hand=None, index=None):
        self.board = parent
        self.hand = hand
        self.index = index
        self.role = -1

    def __repr__(self):
        """Return a string representation of player."""
        return f'player: {self.hand}'

    def make_bid(self, update_bid_history=True):
        """Make a bid and return bid object."""
        active_bid_history = self._active_bid_history(self.board.bid_history)
        self.role = self._get_role(active_bid_history)
        self.board.active_bid_history = active_bid_history

        bid = AcolBid(self.hand, self.board, self.role).bid

        if update_bid_history:
            self.board.bid_history.append(bid.name)
        hc_points = self.hand.high_card_points
        if bid.use_shortage_points:
            distribution_points = 0
            hand_description = f'{hc_points}+{distribution_points} = {hc_points+distribution_points}'
        else:
            hand_description = str(hc_points)
        hand_description = f'{hand_description} '
        bid.hand_points = hand_description
        return bid

    def _get_role(self, bid_history):
        """Return role based on bid history."""
        role_id = self._get_role_id(bid_history)
        if role_id == 0:
            role = ROLES['Opener']
        elif role_id == 2:
            role = ROLES['Responder']
        else:
            role = self._assign_overcaller_advancer(bid_history, role_id)
        return role

    def _get_role_id(self, bid_history):
        """Return the role_id based on the length of the bid_history."""
        role_id = len(bid_history)-100
        for bid in bid_history:
            if bid != 'P':
                break
            else:
                role_id -= 1
        role_id %= self.NUMBER_OF_PLAYERS
        return role_id

    def _assign_overcaller_advancer(self, bid_history, role_id):
        """Return the role_id if Overcaller or Advancer."""
        first_overcaller = 1
        for bid in bid_history[1::2]:
            if bid != 'P':
                break
            else:
                first_overcaller += 2
        first_overcaller %= self.NUMBER_OF_PLAYERS
        if first_overcaller == role_id:
            role = ROLES['Overcaller']
        else:
            role = ROLES['Advancer']
        return role

    @staticmethod
    def _active_bid_history(bid_history):
        """Return the bid history without leading PASSES."""
        temp_history = []
        started = False
        for bid in bid_history:
            if bid != 'P' or started:
                temp_history.append(bid)
                started = True
        return temp_history



""" Bid for Game
    Acol OpenersLaterBid module
"""

import inspect
from .bidding import Bid, Pass
from .blackwood import Blackwood
from .hand import Hand


class OpenersLaterBid(Hand):
    """BfG OpenersLaterBid class."""
    def __init__(self, hand_cards, board):
        super(OpenersLaterBid, self).__init__(hand_cards, board)
        self.my_last_bid = Bid(self.bid_history[-4], '')
        self.my_penultimate_bid = Bid(self.bid_history[-8], '')
        self.partners_last_bid = Bid(self.bid_history[-2], '')
        self.partners_penultimate_bid = Bid(self.bid_history[-6], '')
        self.trace = 0

    def suggested_bid(self):
        """Direct control to relevant method and returns a Bid object."""
        if self.my_last_bid.name == '4NT':
            bid = self.bid_after_4nt()
        elif self.my_last_bid.name == '5NT':
            bid = Blackwood(self.cards, self.board).select_contract()
        elif self.partners_last_bid.name == '4NT':
            bid = Blackwood(self.cards, self.board).count_aces()
        elif self.partners_last_bid.name == '5NT':
            bid = Blackwood(self.cards, self.board).count_kings()
        elif (self.responder_bid_one.denomination != self.responder_bid_two.denomination and
              self.responder_bid_three.denomination == self.opener_bid_one.denomination and
              self.responder_bid_one.denomination.is_major and
              self.suit_length(self.responder_bid_one.denomination) >= 3 and
              not Bid(self.bid_history[-2]).is_game):
            bid = self.bid_to_game(self.responder_bid_one.denomination, '0936')
        elif self._responder_supports_last_bid_suit():
            bid = self.bid_to_game(self.partners_last_bid.denomination, '0536')
        elif (self.hcp >= 23 and
                self.responder_bid_two.denomination.is_suit and
                self.responder_bid_two.denomination != self.responder_bid_one.denomination and
                self.next_level(self.responder_bid_two.denomination) < 6):
            bid = self._strong_and_responder_changes_suit()
        else:
            bid = Pass('0537')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def bid_after_4nt(self):
        """Return bid after response to 4NT."""
        four_aces = Blackwood(self.cards, self.board).four_aces
        if four_aces and self.kings != 4:
            bid = self.nt_bid(5, '0538')
        elif (Blackwood(self.cards, self.board).has_three_aces and
              self.kings == 4 and self._suit_fit()):
            bid = self.suit_bid(6, self._suit_fit(), '0709')
        else:
            bid = Blackwood(self.cards, self.board).select_contract()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _strong_and_responder_changes_suit(self):
        """Return bid after responder changes suit, showing 5/5."""
        if self.suit_length(self.responder_bid_one.denomination) >= 3:
            bid = self.suit_bid(6, self.responder_bid_one.denomination, '0000')
        elif self.suit_length(self.responder_bid_two.denomination) >= 3:
            bid = self.suit_bid(6, self.responder_bid_two.denomination, '0867')
        else:
            bid = self.nt_bid(6, '0000')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid


    # Various utility functions
    def _suit_fit(self):
        """Return the suit if a fit is found."""
        suits = []
        our_bids = [bid for bid in self.bid_history[::-1][1::2]]
        responders_bids = our_bids[::2]
        for bid in responders_bids:
            call = Bid(bid)
            if call.level <= 4 and call.is_suit_call:
                if call.denomination != self.club_suit:
                    suits.append(call.denomination)
                elif call.denomination == self.club_suit and Bid(self.bid_history[0]).is_suit_call:
                    suits.append(call.denomination)
        for suit in suits:
            if self.suit_length(suit) >= 4:
                return suit
        return None

    # Various boolean functions

    def _responder_supports_last_bid_suit(self):
        """Return True if zzz."""
        result = (self.my_last_bid.denomination != self.my_penultimate_bid.denomination and
                  (self.partners_last_bid.denomination == self.my_last_bid.denomination or
                   self.partners_last_bid.denomination == self.my_penultimate_bid.denomination) and
                  self.partners_last_bid.denomination.is_suit and
                  not self.bidding_above_game)
        return result

    # def xxx(self):
    #     """Return True if zzz."""
    #     result = False
    #     return result
    #
    # def xxx(self):
    #     """Return True if zzz."""
    #     result = False
    #     return result
